'''implements functions to calculate statistic informations for a data set

This module contains several functions for the calculation of statistic information about a data set such as minimum, maximum, mean value etc.
The data is used to characterize a data set and to compare them.
 
@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>

@var CHARACTERISTICS_LIST: Is a dictionary with the entry pairs {characteristic_name : function_for_calculating}.
@type CHARACTERISTICS_LIST: dictionary
'''
from numpy import mean, std, load
from scipy.stats import skew, kurtosis, mstats
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA, FOLDER_FOR_PLOTS
from os import path
from time import strftime
import pylab

CHARACTERISTICS_LIST = {'mean value' : lambda x : x.mean(),
                        'standard deviation' : lambda x : x.std(),
                        'skew' : skew,
                        'kurtosis' : kurtosis,
                        'five percent quantile' : lambda x : mstats.mquantiles(x, [0.05])[0],
                        'minimum' : lambda x : x.min(),
                        'maximum' : lambda x : x.max(),
                        'range' : lambda x : x.max()-x.min()}


# dict keys name constraints: lower case, letters only, no special chars
def get_statistic_characteristics(filename):
    """Calculates the characteristics of a data file
    
    This function calculates the characteristics for a data file given gy filename. The path is relative from the default folder of data.
    It returns a dictionary with the characteristics.
    
    @param filename: It is the relative path to the file of which the characteristics should be calculated. The path will be completed by the default folder for data.
    @type filename: string
    @return: Returns a dictionary with the entry pairs {characteristic_name : characteristic_value}.
    @rtype: dictionary
    """  
    result = {}
    data = load(FOLDER_FOR_DATA + '/' + filename)
    for characteristica in CHARACTERISTICS_LIST:
        result[characteristica] = CHARACTERISTICS_LIST[characteristica](data)
    return result


#TODO: Correlation coefficient

# maybe also runtime estimation per sample1
def dynamic_characteristic_leveling(data, samples=0, start_index=0,
                                    characteristic_function=mean):
    """Returns leveled characteristics.
    
    @param data: cost criteria in chronological order (out of optimization)
    @type data: numpy.array
    
    @param samples: maximum of samples considered for leveling
    @type samples: int
    
    @param start_index: index where leveling starts
    @type start_index: int
    
    @param characteristic_function: function of statistical property to investigate
    @type characteristic_function: function-of-numpy
    
    @return: Returns a list with the calculated characteristics.
    @rtype: list-of-float
    
    """
    result = []
    if samples == 0:
        samples = len(data)
    for i in range(samples):
        result.append(characteristic_function(data[start_index:i+start_index+1]))   
    return result

def test_dynamic_leveling(data, samples=2000, cut_away=0, trials=5,
                          characteristic='mean value', final_value=False,
                          sum_stats=False, create_png=False):
    """Plots leveling of the given characteristic.
    Plots the leveling of the given characteristic. X-axis represents the
    number of used samples to achieve the corresponding value. If sum_stats
    is True, the standard deviation and the mean value is calculated over all
    trials for each number of used samples. 
    
    To generate several trials, data is split into multiple parts. Thus, the
    current approach to generate different trials is only applicable for
    Monte Carlo generated date (randomly and independent at each point).
    
    TODO!!! To investigate the leveling for other kind of data a new approach
    is necessary! Until that use trials = 1 in the accordant cases.
    
    @param data: cost criteria in chronological order (out of optimization)
    @type data: numpy.array
    
    @param samples: maximum of samples considered for leveling
    @type samples: int
    
    @param cut_away: number of data points to omit (skip initial stage)
    @type cut_away: int
    
    @param trials: how many different trials are used
    @type trials: int
    
    @param characteristic: statistical property to investigate
    @type characteristic: string
    
    @param final_value: Sets output behavior.
        - True: Enable output the statistical property calculated over the complete data set
        - False: Disable output the statistical property calculated over the complete data set
    @type final_value: boolean
    
    @param sum_stats: Sets trials behavior.
        - True: Enable replacing trials with their mean and std values.
        - False: Disable replacing trials with their mean and std values.
    @type sum_stats: boolean
    
    @param create_png: Sets print behavior.
        - True: Creates a png file in default plot folder.
        - False: Opens interactive mode of matplotlib. 
    @type create_png: boolean
    """
    assert(characteristic in CHARACTERISTICS_LIST.keys())
    characteristic_function = CHARACTERISTICS_LIST[characteristic]
    pylab.interactive(False)
    #pylab.plot([cut_away+1, data.min()])
    #pylab.plot([cut_away+1, data.max()])
    if sum_stats:
        complete_data = []
    if final_value:
        c_final = characteristic_function(data)
        pylab.plot([cut_away+1, samples], [c_final, c_final], c='black', ls='--', lw=1.5)
    for i in range(trials):
        start_index = ((len(data) - samples) // trials) * i
        #print start_index
        print '.', 
        trial_data = dynamic_characteristic_leveling(data, samples, start_index, characteristic_function)
        if sum_stats:
            complete_data.append(trial_data)
        else:
            pylab.plot(range(cut_away+1, samples+1), trial_data[cut_away:])
    if sum_stats:
        # standard deviation of all trials at each sample point
        pylab.subplot(2, 1, 1)
        pylab.plot([std(x) for x in zip(*complete_data)])
        # mean value of all trials at each sample point
        pylab.subplot(2, 1, 2)
        pylab.plot([mean(x) for x in zip(*complete_data)])
    if create_png:
        i = 1
        filename = 'Statistic Characteristic Evolution %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        while path.exists(FOLDER_FOR_PLOTS + '/' + filename):
            i += 1
            filename = 'Statistic Characteristic Evolution %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        pylab.savefig(FOLDER_FOR_PLOTS + '/' + filename)
        pylab.close()
        return filename
    else:
        pylab.show()
