'''implements a function to plot a distribution of two correlated data sets

This module implements a function to plot a distribution of two correlated data sets.
The color show the concentration of correlated solution pairs. It is used to evaluate two cost criteria in their correlated occurrence
and should be used with Monte Carlo searches.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''
import numpy
import pylab
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA, FOLDER_FOR_PLOTS
from time import strftime
from os import path
from scipy import histogram2d

def plot_2d_cost_distribution(filenames, information, parameter, create_png=False):
    
    """Plots two-dimensional cost distribution.
    
    Plots two different data sets against each other (e.g., same benchmark,
    same data structure, but different cost criteria). Thus, it is possible
    to visualize the correlation between this data.
    
    @param filenames: filenames relative to the default folder of data
    @type filenames: list
    @param information: information about the files in filenames extracted from XML data base
    @type information: dictionary
    @param parameter: parameters for the diagram layout
    @type parameter: dictionary
    @param create_png: create a png file(True) from diagram or show interactive mode(False)
    @type create_png: boolean
    """
    grid_size = 500
    #get different dictionaries
    labels = parameter["labels"]
    sizes = parameter["sizes"]
    #axes = parameter["axes"]
    
    #get information about lables
    title = labels["title"]
    #legend_position = labels["legend-position"]
    #legend_title = labels["legend-title"]
    #legend_parameter = labels["legend-parameter"]
    xlabel = labels["x-label"]
    ylabel = labels["y-label"]
    
    #get information about size
    width = sizes["dimensions"]["width"]
    length = sizes["dimensions"]["length"]
    left = sizes["margins"]["left"]
    right = sizes["margins"]["right"]
    top = sizes["margins"]["top"]
    bottom = sizes["margins"]["bottom"]
    
    #===========================================================================
    # if axes["x-axe"]["xmin"] == "auto":
    #    xmin = 0.0
    # else:
    #    xmin = float(axes["x-axe"]["xmin"])
    # if axes["x-axe"]["xmax"] == "auto":
    #    xsearch = True
    #    xmax = 0.0
    # else:
    #    xmax = float(axes["x-axe"]["xmax"])
    #    xsearch = False
    # if axes["y-axe"]["ymin"] == "auto":
    #    ymin = 0.0
    # else:
    #    ymin = float(axes["y-axe"]["ymin"])
    # if axes["y-axe"]["ymax"] == "auto":
    #    ymax = 0.0
    #    ysearch= True 
    # else:
    #    ymax = float(axes["y-axe"]["ymax"])
    #    ysearch = False
    #===========================================================================
    
    #create figure
    pylab.figure(1,(length,width))
    pylab.subplots_adjust(left,bottom,right,top)
    pylab.ticklabel_format(style = "sci", scilimits = (-2,5))
    
    if len(filenames) != 2:
        raise ValueError
    
    horizontal = numpy.load(path.join(FOLDER_FOR_DATA, filenames[0]))
    vertical = numpy.load(path.join(FOLDER_FOR_DATA, filenames[1]))
    H, xedges, yedges = histogram2d(horizontal,
                                    vertical,
                                    bins=grid_size,
                                    normed=True)
    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    pylab.imshow(H, extent=extent, origin='lower')
    
    #set axis names    
    pylab.xlabel(xlabel)
    pylab.ylabel(ylabel)
    
    #set grid
    pylab.grid(True)
    #set title
    pylab.suptitle(title)
    #===========================================================================
    # #create legend
    # pylab.legend(title = legend_title, loc = legend_position)
    #===========================================================================
    if create_png:
        i = 1
        filename = '2D Cost Distribution %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        while path.exists(FOLDER_FOR_PLOTS + '/' + filename):
            i += 1
            filename = '2D Cost Distribution %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        pylab.savefig(FOLDER_FOR_PLOTS + '/' + filename)
        pylab.close()
        return filename
    else:
        pylab.show()
    pass