'''implements a function to plot a histogram of a data set

This module implements a function to plot a histogram of a data set.
It is used to draw the envelope of a cost criteria and should be used with Monte Carlo searches.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''

import numpy
from os import path
from time import strftime
import matplotlib.pyplot as pylab
#from dsc_suite.analyses.standards import SIZES
#from dsc_suite.analyses.standards import AXES

from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA, FOLDER_FOR_PLOTS

def plot_histogram(filenames, information, parameter, create_png = False):
    '''Plots the envelope of the histogram of the data.
    
    Plots the envelope of the histogram of a cost criteria. Useful to compare one benchmark represented by different data structures.
        
    @param filenames: filenames relative to the default folder of data
    @type filenames: list
    @param information: information about the files in filenames extracted from XML data base
    @type information: dictionary
    @param parameter: parameters for the diagram layout
    @type parameter: dictionary
    @param create_png: create a png file(True) from diagram or show interactive mode(False)
    @type create_png: boolean
    '''
    #get different dictionaries
    labels = parameter["labels"]
    sizes = parameter["sizes"]
    axes = parameter["axes"]
    
    #get information about lables
    title = labels["title"]
    legend_position = labels["legend-position"]
    legend_title = labels["legend-title"]
    legend_parameter = labels["legend-parameter"]
    xlabel = labels["x-label"]
    ylabel = labels["y-label"]
    
    #get information about size
    width = sizes["dimensions"]["width"]
    length = sizes["dimensions"]["length"]
    left = sizes["margins"]["left"]
    right = sizes["margins"]["right"]
    top = sizes["margins"]["top"]
    bottom = sizes["margins"]["bottom"]
    
    
    #get information about axes
    if axes["x-axe"]["xmin"] == "auto":
        xmin = 0.0
    else:
        xmin = float(axes["x-axe"]["xmin"])
    if axes["x-axe"]["xmax"] == "auto":
        xsearch = True
        xmax = 0.0
    else:
        xmax = float(axes["x-axe"]["xmax"])
        xsearch = False
    if axes["y-axe"]["ymin"] == "auto":
        ymin = 0.0
    else:
        ymin = float(axes["y-axe"]["ymin"])
    if axes["y-axe"]["ymax"] == "auto":
        ymax = 0.0
        ysearch= True 
    else:
        ymax = float(axes["y-axe"]["ymax"])
        ysearch = False    
    num_bins = 50  
    #create figure
    pylab.figure(1,(length,width))
    pylab.subplots_adjust(left,bottom,right,top)
    pylab.ticklabel_format(style = "sci", scilimits = (-2,5))
    
    arrays = []
    legend_labels = []
    #load data
    for file in filenames:
        array = numpy.load(path.join(FOLDER_FOR_DATA, file))
        arrays.append(array)
        try:
            legend_labels.append(information[file][legend_parameter])
        except:
            try: 
                legend_labels.append(information[file][legend_parameter.replace(" ", "_")])
            except:
                for key in information[file]["cost_criteria"]:
                    try:
                        legend_labels.append(information[file]["cost_criteria"][key][legend_parameter])
                    except:
                        try:
                            legend_labels.append(information[file]["cost_criteria"][key][legend_parameter.replace(" ","_")])
                        except:
                            pass
    #create labels
    if legend_labels != []:
        for i in range(0,len(arrays),1):
            ret = pylab.hist(arrays[i], bins=num_bins, label=legend_labels[i], histtype='step')
            if numpy.amax(ret[1]) > xmax and xsearch:
                xmax = numpy.amax(ret[1])
            if numpy.amax(ret[0]) > ymax and ysearch:
                ymax = numpy.amax(ret[0])*1.01
    else:
        for i in range(0,len(arrays),1):
            ret = pylab.hist(arrays[i], bins=num_bins, histtype='step')
            if numpy.amax(ret[1]) > xmax and xsearch:
                xmax = numpy.amax(ret[1])
            if numpy.amax(ret[0]) > ymax and ysearch:
                ymax = numpy.amax(ret[0])*1.01
    #set title
    pylab.suptitle(title)
    #create legend
    if legend_labels != []:
        pylab.legend(title = legend_title, loc=legend_position)
    #set axis names
    pylab.xlabel(xlabel)
    pylab.ylabel(ylabel)
    #set axis limits --> edit individually 
    pylab.axis([xmin, xmax, ymin, ymax])
    pylab.grid(True, which='both')
    #save png
    if create_png:
        i = 1
        filename = 'Histogram %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        while path.exists(FOLDER_FOR_PLOTS + '/' + filename):
            i += 1
            filename = 'Histogram %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        pylab.savefig(FOLDER_FOR_PLOTS + '/' + filename)
        pylab.close()
        return filename
    pylab.show()