'''

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''
import numpy
from os import path
from time import strftime
import pylab
from scipy import histogram

from dsc_suite.analyses.standards import COLORS
from dsc_suite.gui.config import DATA_STRUCTURE_LIST,BENCHMARK_LIST
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA, FOLDER_FOR_PLOTS

def plot_histogram_operation(filenames, information, parameter, create_png=False):
    '''
    
    @param filenames: filenames relative to the default folder of data
    @type filenames: list
    @param information: information about the files in filenames extracted from XML data base
    @type information: dictionary
    @param parameter: parameters for the diagram layout
    @type parameter: dictionary
    @param create_png: create a png file(True) from diagram or show interactive mode(False)
    @type create_png: boolean
    '''
    if len(filenames) != 2:
        #raise Error
        return None
        pass
    
    #get different dictionaries
    labels = parameter["labels"]
    sizes = parameter["sizes"]
    axes = parameter["axes"]
    
    #get information about lables
    title = labels["title"]
    legend_position = labels["legend-position"]
    legend_title = labels["legend-title"]
    #legend_parameter = labels["legend-parameter"]
    xlabel = labels["x-label"]
    ylabel = labels["y-label"]
    
    #get information about size
    width = sizes["dimensions"]["width"]
    length = sizes["dimensions"]["length"]
    left = sizes["margins"]["left"]
    right = sizes["margins"]["right"]
    top = sizes["margins"]["top"]
    bottom = sizes["margins"]["bottom"]
    
    #get information about axes
    if axes["x-axe"]["xmin"] == "auto":
        xmin_search = True
        xmin = 0.0
    else:
        xmin = float(axes["x-axe"]["xmin"])
        xmin_search = False
    if axes["x-axe"]["xmax"] == "auto":
        xmax_search = True
        xmax = 0.0
    else:
        xmax = float(axes["x-axe"]["xmax"])
        xmax_search = False
    if axes["y-axe"]["ymin"] == "auto":
        ymin = 0.0
        ymin_search = True
    else:
        ymin = float(axes["y-axe"]["ymin"])
        ymin_search = False
    if axes["y-axe"]["ymax"] == "auto":
        ymax = 0.0
        ymax_search= True 
    else:
        ymax = float(axes["y-axe"]["ymax"])
        ymax_search = False
    #create figure
    pylab.figure(1,(length,width))
    pylab.subplots_adjust(left,bottom,right,top)
    pylab.ticklabel_format(style = "sci", scilimits = (-2,5))

    
    #init evaluation
    print_dict = {}
    basic_solutions = 0
    derived_solutions = 0
    data_structure = None
    
    #reading data from datafiles and simplify access via dict
    #reads used benchmark and data structure as well
    #benchmark is necessary for instance of data structure
    #data structure is necessary for operations in evaluation
    for name in filenames:
        dictionary = information[name]["cost_criteria"]
        benchmark = BENCHMARK_LIST[information[name]["benchmark"]][0](information[name]["benchmark"])
        benchmark.to_3d()
        data_structure = DATA_STRUCTURE_LIST[information[name]["datastructure"]][0](benchmark)
        for key in dictionary.keys():
            if str(dictionary[key]["filename"]) == str(name):
                print_dict.update({key : numpy.load(path.join(FOLDER_FOR_DATA, name))})
    
    #without ITER file evaluation is not possible
    #leads to an exception in the main program
    if "ITER" not in print_dict.keys():
        return None
    
    #remove ITER file from plotting data access
    criteria = print_dict.keys()
    criteria.remove("ITER")
    
    #create lists for plotting data
    basics_x = []
    difference_x = []
    
    #number of bins for the histogram -> adjustable only here
    num_bins = 100
    
    #getting operations 
    ops = data_structure.get_operations()
    op_names = ops.keys()
    op_names.sort()
    
    #create lists for data for all operations
    for i in range(0,len(op_names),1):
        difference_x.append([])
    
    #separate data into basics and derived solutions
    #separation uses ITER index
    for i in range(0,len(print_dict["ITER"]),1):
        if print_dict["ITER"][i].startswith("b-"):
            basics_x.append(print_dict[criteria[0]][i])
            
        elif print_dict["ITER"][i].startswith("d-"):
            code = print_dict["ITER"][i].split("-")
            if int(code[4]) != 0:
                continue
            save_index, base_index = int(code[1]),int(code[2])
            difference = print_dict[criteria[0]][i]-basics_x[base_index]
            difference_x[save_index].append(difference)
    
    #calculate statistic data
    mean_values = []
    deviations  = []
    
    for array in difference_x:
        mean_values.append(numpy.mean(array))
        deviations.append(numpy.std(array))
    
    avg_mean = sum(mean_values)/len(mean_values)
    avg_sigma = sum(deviations)/len(deviations)
    
    if xmax_search:
        xmax = avg_mean+3*avg_sigma
    if xmin_search:
        xmin = avg_mean-3*avg_sigma
         
    step_size = (6*avg_sigma)/num_bins
    calculated_bins = pylab.frange(avg_mean-3*avg_sigma, avg_mean+3*avg_sigma, step_size)
           
    for name in op_names:
        histo, bins = histogram(difference_x[op_names.index(name)], calculated_bins)
        histo = numpy.float64(histo)
        histo *= 1.0/len(difference_x[op_names.index(name)])
        if numpy.amax(histo) > ymax and ymax_search:
            ymax = numpy.amax(histo)
        bincenters = 0.5*(bins[1:] + bins[:-1])
        pylab.plot(bincenters, histo, '-', alpha=0.7, color=COLORS[op_names.index(name)], label=name)
    
    #set axis names    
    pylab.xlabel(xlabel)
    pylab.ylabel(ylabel)
    #set axis limits
    pylab.axis([xmin, xmax, ymin, ymax])
    #set grid
    pylab.grid(True)
    #set title
    pylab.suptitle(title)
    #create legend --> color map
    pylab.legend(title = legend_title, loc = legend_position)
    #create png
    if create_png:
        i = 1
        filename = 'Histogram Operation %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        while path.exists(FOLDER_FOR_PLOTS + '/' + filename):
            i += 1
            filename = 'Histogram Operation %s %02i.png' % (strftime("%Y-%m-%d %H%M%S"), i)
        pylab.savefig(FOLDER_FOR_PLOTS + '/' + filename)
        pylab.close()
        return filename
    else:
        pylab.show()
