"""implements the Sequence Quintuple data structure

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
"""
from dsc_suite.ds.data_structure import DataStructure
from dsc_suite.tools.combinatorics import EnumerativeCombinatorics
from dsc_suite.tools.graphtheory import longest_path
from dsc_suite.tools.geometry import overlap1d
from random import randint


##### SEQUENCE TRIPLE SUBCLASS #################################################
class SequenceQuintuple(DataStructure):
    """Implements Sequence Quintuple data structure.

    Three sequences used similar as sequence pair. Published by the same group
    in 2000 (Yamazaki2000).

    based on five sequences, so called locii
    sequence_quintuple format : [locii_1, locii_2, locii_3, locii_4, locii_5, block_rotations]
        
    TODO: detailed docstring
    """
    name = 'Sequence Quintuple'

    def __init__(self, benchmark):
        """Creates a new instance.
        
        @param benchmark: A benchmark instance of a loaded benchmark.
        @type Benchmark
        """
        assert(benchmark.status['number of dimensions'] == 3)
        DataStructure.__init__(self, benchmark)
        from operator import mul
        from dsc_suite.tools.math import fak
        # maybe better combined in a status dictionary
        self.num_blocks = len(benchmark.blocks)
        self.num_parts = 5 + 1
        self.part_lengths = [fak(self.num_blocks)]*5 + [2**self.num_blocks]
        self.num_representations = reduce(mul, self.part_lengths)
        # if constructor call takes to long comment out
        #self.update_balancing()
    

    def get_all_representations(self):
        """Generates all possible sequence quintuples.
        This method generates all possible sequence quintuples.
        len(abstract_solutions) = n!^5, where n is the number of modules
        
        @return: Returns a list of all possible abstract solutions.
        @rtype: list
        """
        locii_1 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        locii_2 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        locii_3 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        locii_4 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        locii_5 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        rotations = EnumerativeCombinatorics.vary([self.ORIGINAL, self.ROTATED],
                                                  len(self.benchmark.blocks))
        representations = []
        for L1 in locii_1:
            for L2 in locii_2:
                for L3 in locii_3:
                    for L4 in locii_4:
                        for L5 in locii_5:
                            for r in rotations: 
                                representations.append([L1,L2,L3,L4,L5,r])
        return representations
    
    def get_part_of_representation(self, part_index, part_number):
        """Gets a single element out of a part of the abstract representation.
        part_number between 0 and part_lengths[part_index]
        
        @param part_index:
        @type part_index: int
        
        @param part_number:
        @type part_number: int
        """
        # parameter part_index is not used here because all three
        # sequences are equal, but for other data structures or by adding
        # a rotation sequence, this parameter is useful
        if 0 <= part_index <= 4:
            return EnumerativeCombinatorics.get_permutation(self.benchmark.blocks, part_number)
        elif part_index == 5:
            return EnumerativeCombinatorics.get_variation([self.ORIGINAL, self.ROTATED],
                                                          len(self.benchmark.blocks), part_number)

    # geometric relations
    # x-direction, a left to b, b right to a
    def left_to(self, sequence_quintuple, a, b,
                x_distance=None, y_distance=None, widths=None, heights=None):
        """
        @param sequence_quintuple:
        @type sequence_quintuple:
        
        @param a:
        @type a:
        
        @param b:
        @type b:
        
        @param x_distance:
        @type x_distance:
        
        @param y_distance:
        @type y_distance:
        
        @param widths:
        @type widths:
        
        @param heights:
        @type heights:
        
        @return:
        @rtype: boolean
        """
        locii1, locii2 = sequence_quintuple[0:2]
        if ((locii1.index(a) < locii1.index(b)) and
            (locii2.index(a) < locii2.index(b))):
            return True
        else:
            return False

    # y-direction, a in front of b, b behind a
    def in_front_of(self, sequence_quintuple, a, b,
                    x_distance=None, y_distance=None, widths=None, heights=None):
        """
        @param sequence_quintuple:
        @type sequence_quintuple:
        
        @param a:
        @type a:
        
        @param b:
        @type b:
        
        @param x_distance:
        @type x_distance:
        
        @param y_distance:
        @type y_distance:
        
        @param widths:
        @type widths:
        
        @param heights:
        @type heights:
        
        @return:
        @rtype: boolean
        """
        locii3, locii4 = sequence_quintuple[2:4]
        if ((locii3.index(a) < locii3.index(b)) and
            (locii4.index(a) < locii4.index(b))):
            return True
        else:
            return False


    # z-direction, a below b, b above a
    def below(self, sequence_quintuple, a, b,
              x_distance=None, y_distance=None, widths=None, heights=None):
        """
        @param sequence_quintuple:
        @type sequence_quintuple:
        
        @param a:
        @type a:
        
        @param b:
        @type b:
        
        @param x_distance:
        @type x_distance:
        
        @param y_distance:
        @type y_distance:
        
        @param widths:
        @type widths:
        
        @param heights:
        @type heights:
        
        @return:
        @rtype: boolean
        """
        locii5 = sequence_quintuple[4]
        if ((locii5.index(a) < locii5.index(b)) and
            overlap1d(x_distance[a]-widths[a], x_distance[a],
                    x_distance[b]-widths[b], x_distance[b]) and
            overlap1d(y_distance[a]-heights[a], y_distance[a],
                    y_distance[b]-heights[b], y_distance[b])):
            return True
        else:
            return False

    
    # i need to construct the constraint graphs
    # the following constructs a constraint graph with transient edges
    def get_constraint_graph(self, sequence_quintuple, relation_function,
                             x_distance=None, y_distance=None, widths=None, heights=None):
        """
        @param sequence_quintuple:
        @type sequence_quintuple:
        
        @param relation_function:
        @type relation_function:
        
        @param x_distance:
        @type x_distance:
        
        @param y_distance:
        @type y_distance:
        
        @param widths:
        @type widths:
        
        @param heights:
        @type heights:
        
        @return:
        @rtype:
        """
        cg = dict()
        children = set()
        for a in self.benchmark.blocks:
            cg[a] = []
            b_list = list(self.benchmark.blocks)
            b_list.remove(a)
            for b in b_list:
                if relation_function(sequence_quintuple, a, b, 
                                     x_distance, y_distance, widths, heights):
                    cg[a].append(b)
                    children.add(b)
        # TODO: Attention! If Source is a module name it will overwriten.
        cg['Source'] = list(set(self.benchmark.blocks) - children)      
        return cg

    # depth, z-direction
    def get_dcg(self, sequence_triple):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @return:
        @rtype:
        """
        return self.get_constraint_graph(sequence_triple, self.below)

    # vielleicht doch lieber solution als parameter und dann
    # sequence_triple und rotations, ... in ein tupel packen
    # flexible (soft) blocks, rectilinear
    def create_packing(self, representation):
        """Returns packing in dictionary format.
        packing = {'name' : [x, y, z, w, h, d], ...}
        
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the created packing.
        @rtype: dictionary
        """
        sequence_quintuple = representation[:5]
        rotations = representation[5]
        dims = self.benchmark.dimensions
        widths = dict(zip(dims.keys(), zip(*dims.values())[0]))
        heights = dict(zip(dims.keys(), zip(*dims.values())[1]))
        depths = dict(zip(dims.keys(), zip(*dims.values())[2]))
        for i in range(len(rotations)):
            if rotations[i] == self.ROTATED:
                block = self.benchmark.blocks[i]
                widths[block], heights[block] = heights[block], widths[block]
        x_distance, x_parent = longest_path(self.get_constraint_graph(sequence_quintuple, self.left_to), widths)
        y_distance, y_parent = longest_path(self.get_constraint_graph(sequence_quintuple, self.in_front_of), heights)
        z_distance, z_parent = longest_path(self.get_constraint_graph(sequence_quintuple, self.below,
                                                                      x_distance, y_distance,
                                                                      widths, heights), depths)
        packing = dict()
        for b in self.benchmark.blocks:
            packing[b] = [x_distance[b]-widths[b],
                          y_distance[b]-heights[b],
                          z_distance[b]-depths[b],
                          widths[b], heights[b], depths[b]]
        # TODO: Currently unused property of Sequence Triple, that while
        #       generating longest path, the outer dimensions are already
        #       obtained. This could simplify the cost evaluation.
        return packing

    def packing_to_representation(self, packing):
        """Creates a representation out of a packing.
        
        @param packing: a packing of a current representation
        @type packing: dictionary
        
        @note: Not implemented and/or available!
        """
        print "Not implemented and/or available!"

   
    def get_operations(self):
        """Returns a dictionary of possible operations onto a abstract solution.
        The higher the globality_factor, the bigger is the solution change
        This factor is given by the authors but it also can be determined
        in test runs evaluation the cost influences caused by applying the
        according function.
        
        @return: Returns a dictionary of possible operations onto a abstract solution.
        structure: {'name' : [function, globality_factor]}
        @rtype: dictionary
        """
        return {'Exchange 2 labels in 1 locii' : [self.exchange_two_labels_in_one_locii, 8],
                'Exchange 2 labels in all locii' : [self.exchange_two_labels_in_all_locii, 4],
                'Exchange 2 positions in all locii' : [self.exchange_two_positions_in_all_locii, 10],
                'Exchange 2 locii' : [self.exchange_two_locii, 10],
                'Rotate single block in z-plane': [self.rotate_single_block, 2]}

    
    def merge_representations(self, representation_a, representation_b):
        """Merges two representations into all possible children.
        Typically, used for genetic and/or evolutionary algorithms.
        Returns a list with all possible children.
        
        In case of Sequence Quintuple all possible combinations of the
        first, second, third, fourth and fifth sequence as well as the
        rotation_list are generated. Rotations are currently completely
        taken from representation a or b, respectively.
        Thus 2^6 == 64 different children are possible.
        
        @param representation_a: a representation list
        @type representation_a: list
        
        @param representation_b: a representation list
        @type representation_b: list
        
        @return: Returns a list with all possible representations created by merging the two given.
        @rtype: list
        """
        result = []
        for n in range(2**6):
            child = []
            #print ''
            for i in range(6):
                if n // 2**(5-i):
                    #print '0',
                    child.append(representation_a[i])
                else:
                    #print '1',
                    child.append(representation_b[i])
                n = n % 2**(5-i)
            result.append(child)
        return result

    
    def exchange_two_labels_in_one_locii(self, representation):
        """Exchanges two labels in one locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        choice = randint(0,4)
        sq = list(representation[:5])
        # old locii
        locii = sq[choice]
        pos1 = pos2 = randint(0, len(locii)-1)
        while pos1 == pos2:
            pos2 = randint(0, len(locii)-1)
        if pos1 > pos2:
            pos1, pos2 = pos2, pos1
        # new locii
        sq[choice] = locii[:pos1]+[locii[pos2]]+locii[pos1+1:pos2]+[locii[pos1]]+locii[pos2+1:]
        return sq+[representation[5]]

    def exchange_two_labels_in_all_locii(self, representation):
        """Exchange two labels in all locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        sq = list(representation[:5])
        lab1 = lab2 = sq[0][randint(0, len(sq[0])-1)]
        while lab1 == lab2:
            lab2 = sq[0][randint(0, len(sq[0])-1)]
        for i in range(5):
            # old locii
            locii = sq[i]
            pos1 = locii.index(lab1)
            pos2 = locii.index(lab2)
            if pos1 > pos2:
                pos1, pos2 = pos2, pos1
            # new locii
            sq[i] = locii[:pos1]+[locii[pos2]]+locii[pos1+1:pos2]+[locii[pos1]]+locii[pos2+1:]
        return sq+[representation[5]]
  
    def exchange_two_positions_in_all_locii(self, representation):
        """Exchange two positions in all locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        sq = list(representation[:5])
        pos1 = pos2 = randint(0, len(sq[0])-1)
        while pos1 == pos2:
            pos2 = randint(0, len(sq[0])-1)
        if pos1 > pos2:
            pos1, pos2 = pos2, pos1
        for i in range(5):
            # old locii
            locii = sq[i]
            # new locii
            sq[i] = locii[:pos1]+[locii[pos2]]+locii[pos1+1:pos2]+[locii[pos1]]+locii[pos2+1:]
        return sq+[representation[5]]

    def exchange_two_locii(self, representation):
        """Exchanges two locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        sq = list(representation[:5])
        choice1 = choice2 = randint(0,4)
        while choice1 == choice2:
            choice2 = randint(0, 4)
        temp = sq[choice1]
        sq[choice1] = sq[choice2]
        sq[choice2] = temp
        return sq+[representation[5]]

    def rotate_single_block(self, representation):
        """Rotates single block.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        rotations = list(representation[5])
        index = randint(0, len(rotations)-1)
        if rotations[index] == self.ORIGINAL:
            rotations[index] = self.ROTATED
        else:
            rotations[index] = self.ORIGINAL
        return representation[:5]+[rotations]

    
