"""implements the Sequence Triple data structure

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
"""
from dsc_suite.ds.data_structure import DataStructure
from dsc_suite.tools.combinatorics import EnumerativeCombinatorics
from random import randint


##### SEQUENCE TRIPLE SUBCLASS #################################################
class SequenceTriple(DataStructure):
    """Implements Sequence Triple data structure.

    Three sequences used similar as sequence pair. Published by the same group
    in 2000 (Yamazaki2000).

    based on three sequences, so called locii
    sequence_triple format : [locii_1, locii_2, locii_3, block_rotations]
        
    TODO: detailed docstring
    """
    name = 'Sequence Triple'

    def __init__(self, benchmark):
        """Creates a new instance.
        
        @param benchmark: A benchmark instance of a loaded benchmark.
        @type Benchmark
        """
        assert(benchmark.status['number of dimensions'] == 3)
        DataStructure.__init__(self, benchmark)
        from operator import mul
        from dsc_suite.tools.math import fak
        # maybe better combined in a status dictionary
        self.num_blocks = len(benchmark.blocks)
        self.num_parts = 3 + 1
        self.part_lengths = [fak(self.num_blocks)]*3 + [2**self.num_blocks]
        self.num_representations = reduce(mul, self.part_lengths)
        # if constructor call takes to long comment out
        #self.update_balancing()
    

    def get_all_representations(self):
        """Generates all possible sequence triples.
        This method generates all possible sequence triples.
        len(abstract_solutions) = n!^3, where n is the number of modules
        
        @return: Returns a list of all possible abstract solutions.
        @rtype: list
        """
        locii_1 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        locii_2 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        locii_3 = EnumerativeCombinatorics.permute(self.benchmark.blocks)
        rotations = EnumerativeCombinatorics.vary([self.ORIGINAL, self.ROTATED],
                                                  len(self.benchmark.blocks))
        representations = []
        for x in locii_1:
            for y in locii_2:
                for z in locii_3:
                    for r in rotations:
                        representations.append([x,y,z,r])
        return representations
    
    def get_part_of_representation(self, part_index, part_number):
        """Gets a single element out of a part of the abstract representation.
        part_number between 0 and part_lengths[part_index]
        
        @param part_index:
        @type part_index: int
        
        @param part_number:
        @type part_number: int
        """
        # parameter part_index is not used here because all three
        # sequences are equal, but for other data structures or by adding
        # a rotation sequence, this parameter is useful
        if part_index == 0 or part_index == 1 or part_index == 2:
            return EnumerativeCombinatorics.get_permutation(self.benchmark.blocks, part_number)
        elif part_index == 3:
            return EnumerativeCombinatorics.get_variation([self.ORIGINAL, self.ROTATED],
                                                          len(self.benchmark.blocks), part_number)

    # geometric relations
    # y-direction, a in front of b, b behind a
    def in_front_of(self, sequence_triple, a, b):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @param a:
        @type a:
        
        @param b:
        @type b:

        @return:
        @rtype: boolean
        """
        locii1, locii2, locii3 = sequence_triple
        if ((locii1.index(a) < locii1.index(b)) and
            (locii2.index(a) < locii2.index(b)) and
            (locii3.index(a) < locii3.index(b))):
            return True
        else:
            return False

    # x-direction, a left to b, b right to a
    def left_to(self, sequence_triple, a, b):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @param a:
        @type a:
        
        @param b:
        @type b:

        @return:
        @rtype: boolean
        """
        locii1, locii2, locii3 = sequence_triple
        if (((locii1.index(a) < locii1.index(b)) and
             (locii2.index(a) > locii2.index(b)) and
             (locii3.index(a) < locii3.index(b))) or
            ((locii1.index(a) > locii1.index(b)) and
             (locii2.index(a) > locii2.index(b)) and
             (locii3.index(a) < locii3.index(b)))):
            return True
        else:
            return False

    # z-direction, a below b, b above a
    def below(self, sequence_triple, a, b):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @param a:
        @type a:
        
        @param b:
        @type b:

        @return:
        @rtype: boolean
        """
        locii1, locii2, locii3 = sequence_triple
        if ((locii1.index(a) > locii1.index(b)) and
            (locii2.index(a) < locii2.index(b)) and
            (locii3.index(a) < locii3.index(b))):
            return True
        else:
            return False
    
    # i need to construct the constraint graphs
    # the following constructs a constraint graph with transient edges
    def get_constraint_graph(self, sequence_triple, relation_function):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @param relation_function:
        @type relation_function:

        @return:
        @rtype:
        """
        cg = dict()
        children = set()
        for a in self.benchmark.blocks:
            cg[a] = []
            b_list = list(self.benchmark.blocks)
            b_list.remove(a)
            for b in b_list:
                if relation_function(sequence_triple, a, b):
                    cg[a].append(b)
                    children.add(b)
        # TODO: Attention! If Source is a module name it will overwriten.
        cg['Source'] = list(set(self.benchmark.blocks) - children)      
        return cg

    # horizontal, x-direction
    def get_hcg(self, sequence_triple):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @return:
        @rtype:
        """
        return self.get_constraint_graph(sequence_triple, self.left_to)

    # vertical, y-direction
    def get_vcg(self, sequence_triple):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @return:
        @rtype:
        """
        return self.get_constraint_graph(sequence_triple, self.in_front_of)

    # depth, z-direction
    def get_dcg(self, sequence_triple):
        """
        @param sequence_triple:
        @type sequence_triple:
        
        @return:
        @rtype:
        """
        return self.get_constraint_graph(sequence_triple, self.below)

    # vielleicht doch lieber solution als parameter und dann
    # sequence_triple und rotations, ... in ein tupel packen
    # flexible (soft) blocks, rectilinear
    def create_packing(self, representation):
        """Returns packing in dictionary format.
        packing = {'name' : [x, y, z, w, h, d], ...}
        
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the created packing.
        @rtype: dictionary
        """
        from dsc_suite.tools.graphtheory import longest_path
        sequence_triple = representation[:3]
        rotations = representation[3]
        dims = self.benchmark.dimensions
        widths = dict(zip(dims.keys(), zip(*dims.values())[0]))
        heights = dict(zip(dims.keys(), zip(*dims.values())[1]))
        depths = dict(zip(dims.keys(), zip(*dims.values())[2]))
        for i in range(len(rotations)):
            if rotations[i] == self.ROTATED:
                block = self.benchmark.blocks[i]
                widths[block], heights[block] = heights[block], widths[block]
        x_distance, x_parent = longest_path(self.get_hcg(sequence_triple), widths)
        y_distance, y_parent = longest_path(self.get_vcg(sequence_triple), heights)
        z_distance, z_parent = longest_path(self.get_dcg(sequence_triple), depths)
        packing = dict()
        for b in self.benchmark.blocks:
            packing[b] = [x_distance[b]-widths[b],
                          y_distance[b]-heights[b],
                          z_distance[b]-depths[b],
                          widths[b], heights[b], depths[b]]
        # TODO: Currently unused property of Sequence Triple, that while
        #       generating longest path, the outer dimensions are already
        #       obtained. This could simplify the cost evaluation.
        return packing

    def packing_to_representation(self, packing):
        """Creates a representation out of a packing.
        
        @param packing: a packing of a current representation
        @type packing: dictionary
        
        @note: Not implemented and/or available!
        """
        print "Not implemented and/or available!"

    def get_operations(self):
        """Returns a dictionary of possible operations onto a abstract solution.
        The higher the globality_factor, the bigger is the solution change
        This factor is given by the authors but it also can be determined
        in test runs evaluation the cost influences caused by applying the
        according function.
        
        @return: Returns a dictionary of possible operations onto a abstract solution.
        structure: {'name' : [function, globality_factor]}
        @rtype: dictionary
        """
        return {'Exchange 2 labels in 1 locii' : [self.exchange_two_labels_in_one_locii, 8],
                'Exchange 2 labels in all locii' : [self.exchange_two_labels_in_all_locii, 4],
                'Exchange 2 positions in all locii' : [self.exchange_two_positions_in_all_locii, 10],
                'Exchange 2 locii' : [self.exchange_two_locii, 10],
                'Rotate single block in z-plane': [self.rotate_single_block, 2]}

    def merge_representations(self, representation_a, representation_b):
        """Merges two representations into all possible children.
        Typically, used for genetic and/or evolutionary algorithms.
        Returns a list with all possible children.
        
        In case of Sequence Triple all possible combinations of the
        first, second and third sequence as well as the rotation_list
        are generated. Rotations are currently completely taken from
        representation a or b, respectively. Thus 2^4 = 16 different children
        are possible.
        
        @param representation_a: a representation list
        @type representation_a: list
        
        @param representation_b: a representation list
        @type representation_b: list
        
        @return: Returns a list with all possible representations created by merging the two given.
        @rtype: list
        """
        result = []
        for n in range(2**4):
            child = []
            for i in range(4):
                if n // 2**(3-i):
                    child.append(representation_a[i])
                else:
                    child.append(representation_b[i])
                n = n % 2**(3-i)
            result.append(child)
        return result
    
    def exchange_two_labels_in_one_locii(self, representation):
        """Exchanges two labels in one locii.
        @param representation: a representation list
        @type representation: list
        """
        choice = randint(0,2)
        st = list(representation[:3])
        # old locii
        locii = st[choice]
        pos1 = pos2 = randint(0, len(locii)-1)
        while pos1 == pos2:
            pos2 = randint(0, len(locii)-1)
        if pos1 > pos2:
            pos1, pos2 = pos2, pos1
        # new locii
        st[choice] = locii[:pos1]+[locii[pos2]]+locii[pos1+1:pos2]+[locii[pos1]]+locii[pos2+1:]
        return st+[representation[3]]

    def exchange_two_labels_in_all_locii(self, representation):
        """Exchanges two labels in all locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        st = list(representation[:3])
        lab1 = lab2 = st[0][randint(0, len(st[0])-1)]
        while lab1 == lab2:
            lab2 = st[0][randint(0, len(st[0])-1)]
        for i in range(3):
            # old locii
            locii = st[i]
            pos1 = locii.index(lab1)
            pos2 = locii.index(lab2)
            if pos1 > pos2:
                pos1, pos2 = pos2, pos1
            # new locii
            st[i] = locii[:pos1]+[locii[pos2]]+locii[pos1+1:pos2]+[locii[pos1]]+locii[pos2+1:]
        return st+[representation[3]]
  
    def exchange_two_positions_in_all_locii(self, representation):
        """Exchanges two positions in all locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        st = list(representation[:3])
        pos1 = pos2 = randint(0, len(st[0])-1)
        while pos1 == pos2:
            pos2 = randint(0, len(st[0])-1)
        if pos1 > pos2:
            pos1, pos2 = pos2, pos1
        for i in range(3):
            # old locii
            locii = st[i]
            # new locii
            st[i] = locii[:pos1]+[locii[pos2]]+locii[pos1+1:pos2]+[locii[pos1]]+locii[pos2+1:]
        return st+[representation[3]]

    def exchange_two_locii(self, representation):
        """Exchanges two locii.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        st = list(representation[:3])
        choice1 = choice2 = randint(0,2)
        while choice1 == choice2:
            choice2 = randint(0, 2)
        temp = st[choice1]
        st[choice1] = st[choice2]
        st[choice2] = temp
        return st+[representation[3]]
       

    def rotate_single_block(self, representation):
        """Rotates single block.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        rotations = list(representation[3])
        index = randint(0, len(rotations)-1)
        if rotations[index] == self.ORIGINAL:
            rotations[index] = self.ROTATED
        else:
            rotations[index] = self.ORIGINAL
        return representation[:3]+[rotations]

    
