"""implements the T-Tree data structure [Yuh2004a, Yuh2009]

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
"""
from dsc_suite.ds.data_structure import DataStructure
from dsc_suite.tools.combinatorics import EnumerativeCombinatorics as EC
from dsc_suite.tools.graphtheory import TernaryTrees as TT
from dsc_suite.tools.geometry import overlap2d
from random import randrange, randint


##### T-TREE SUBCLASS ####################################################
class TTree(DataStructure):
    """Implements T-Tree data structure.

    t_tree format : [module_permutation, topology, module_rotations]
    
    ternary tree topology is in a list format where each entry
    represents a node. The index of the root node is also returned.        
    panholzer = ([[LEFT, MIDDLE, RIGHT, PARENT, INTERNAL, DEPTH], ...], root)
    topology = TT.transform_panholzer_to_topology(T, root)
    
    rotations only 0 and 90 degrees    
    
    TODO: detailed docstring, verification and rectification step needed after applying permutation operations 
    """
    name = 'T-Tree'
    
    # H data constants
    __H_NAME = 0
    __H_PARENT = 1
    __H_LEFT_CHILD = 2
    __H_MIDDLE_CHILD = 3
    __H_RIGHT_CHILD = 4
    __H_WIDTH = 5
    __H_HEIGHT = 6
    __H_DEPTH = 7
    __H_X = 8
    __H_Y = 9
    __H_Z = 10

    def __init__(self, benchmark):
        """Creates a new instance.
        
        @param benchmark: A benchmark instance of a loaded benchmark.
        @type Benchmark
        """
        assert(benchmark.status['number of dimensions'] == 3)
        DataStructure.__init__(self, benchmark)
        from operator import mul
        self.num_blocks = len(benchmark.blocks)
        self.num_slices = self.num_blocks - 1
        self.num_parts = 2 + 1
        self.part_lengths = [EC.get_num_permutations(benchmark.blocks),
                             TT.get_num_ternary_trees(self.num_blocks),
                             2**self.num_blocks]
        self.num_representations = reduce(mul, self.part_lengths)
        #self.update_balancing()

    def __UpdateContour(self, c, x1, x2, height):
        """Inserts into horizontal contour
        
        @param c: contour, initially c = [[0,0]]
        @type c: list
        
        @param x1:
        @type x1:
        
        @param x2:
        @type x2:
        
        @param height:
        @type height:
        
        @return:
        @rtype:
        """
        temp = list(c)
        temp.sort()
        y = 0
        y_old = -1
        corner = False
        for i in temp:
            if i[0] >= x1 and i[0] < x2:
                y_old = i[1]
                y = max(y, y_old)
                c.remove(i)
            elif i[0] == x2:
                y_old = i[1]
                c.remove(i)
            if i[0] == x1:
                corner = True
        c.append([x1, y+height])
        c.append([x2, y_old])
        c.sort()
        i = c.index([x1, y+height])
        if i > 0 and not corner:
            y = max(y, c[i-1][1])
            c[i][1] = y+height
        if y_old < 0:
            c[i+1][1] = c[i-1][1]
        if c[i][1] == c[i+1][1]:
            c.remove(c[i+1])
        if i > 0 and c[i-1][1] == c[i][1]:
            c.remove(c[i])
        return y
    
    def __TreeDecomposition(self, n, S, H):
        """
        @param n: a node of a binary tree
        @type n:
        
        @param S: a stack
        @type S:
        
        @param H: a node of T-Tree
        @type H:
        
        @return:
        @rtype:
        """
        S_ = []
        b_root = n
        S_.append(n)
        while len(S_) > 0:
            m = S_.pop()
            m_mc = H[m][self.__H_MIDDLE_CHILD]
            m_lc = H[m][self.__H_LEFT_CHILD]
            m_rc = H[m][self.__H_RIGHT_CHILD]
            if not m_mc == None:
                S_.append(m_mc)
            if not m_lc == None:
                S_.append(m_lc)
            if not m_rc == None:
                S.append(m_rc)
        return b_root
    
    def __FindMaxX(self, n, L, H):
        """
        @param n: a node of a binary tree
        @type n:
        
        @param L: a list to store nodes
        @type L: list
        
        @param H: a node of T-Tree
        @type H:
        
        @return:
        @rtype:
        """
        temp = filter(lambda m: overlap2d(H[n][self.__H_Y], H[n][self.__H_Y] + H[n][self.__H_HEIGHT],
                                          H[n][self.__H_Z], H[n][self.__H_Z] + H[n][self.__H_DEPTH],
                                          H[m][self.__H_Y], H[m][self.__H_Y] + H[m][self.__H_HEIGHT],
                                          H[m][self.__H_Z], H[m][self.__H_Z] + H[m][self.__H_DEPTH]), L)
        temp = map(lambda m: H[m][self.__H_X] + H[m][self.__H_WIDTH], temp) + [0]
        return max(temp)
        
    def __PlaceModule(self, n, L, H, contour):
        """
        @param n: a node of a binary tree
        @type n:
        
        @param L: a list to store nodes
        @type L: list
        
        @param H: a node of T-Tree
        @type H:
        
        @param contour:
        @type contour:
        """
        if H[n][self.__H_PARENT] == None:
            H[n][self.__H_Z] = 0
            H[n][self.__H_Y] = self.__UpdateContour(contour, 0,
                                                    H[n][self.__H_WIDTH],
                                                    H[n][self.__H_HEIGHT])
            H[n][self.__H_X] = 0
            L.append(n)
            return
        # n is left child
        parent = H[n][self.__H_PARENT]
        if H[parent][self.__H_LEFT_CHILD] == n:
            H[n][self.__H_Z] = H[parent][self.__H_Z] + H[parent][self.__H_DEPTH]
        else:
            H[n][self.__H_Z] = H[parent][self.__H_Z]
        H[n][self.__H_Y] = self.__UpdateContour(contour,
                                                H[n][self.__H_Z],
                                                H[n][self.__H_Z] + H[n][self.__H_DEPTH],
                                                H[n][self.__H_HEIGHT])
        H[n][self.__H_X] = self.__FindMaxX(n, L, H) 
        L.append(n)
        
    
    def __BinaryTreePacking(self, b_root, H, L):
        """
        @param b_root:
        @type b_root:
        
        @param H: a node of T-Tree
        @type H:
        
        @param L: a list to store nodes
        @type L: list
        """
        S = []
        p = b_root
        if H[p][self.__H_PARENT] == None:
            contour = [[0,0]]
        else:
            contour = [[0,H[H[p][self.__H_PARENT]][self.__H_Y]]]
        self.__PlaceModule(p, L, H, contour)
        if not H[p][self.__H_MIDDLE_CHILD] == None:
            S.append(H[p][self.__H_MIDDLE_CHILD])
        if not H[p][self.__H_LEFT_CHILD] == None:
            S.append(H[p][self.__H_LEFT_CHILD])
        while len(S) > 0:
            p = S.pop()
            self.__PlaceModule(p, L, H, contour)
            if not H[p][self.__H_MIDDLE_CHILD] == None:
                S.append(H[p][self.__H_MIDDLE_CHILD])
            if not H[p][self.__H_LEFT_CHILD] == None:
                S.append(H[p][self.__H_LEFT_CHILD])

        
    def __TreePacking(self, H):
        """ Packs the Tree. 
        Packing process described in [Yuh2009]
        @param H: a T-Tree
        @type H:
        """
        S = []
        L = []
        b_root = self.__TreeDecomposition(0, S, H)
        self.__BinaryTreePacking(b_root, H, L)
        while len(S) > 0:
            n = S.pop()
            b_root = self.__TreeDecomposition(n, S, H)
            self.__BinaryTreePacking(b_root, H, L)

    def __CreateH(self, t_tree, dimensions):
        """Creates the DFS list structure used by t-tree packing algorithm.
        H = [[name, parent, left, middle, right, width, height, depth, x, y, z], ...]
        @param t_tree: a T-Tree
        @type t_tree:
        
        @param dimensions:
        @type dimensions:
        """
        module_permutation, topology, module_rotations = t_tree
        H = []
        for i in range(len(topology)):
            name = module_permutation[i]
            parent, left_child, middle_child, right_child = topology[i]
            width, height, depth = dimensions[name]
            if module_rotations[i] == 1:
                width, height = height, width
            H.append([name, parent, left_child, middle_child, right_child,
                      width, height, depth, 0, 0, 0])
        return H

    def create_packing(self, representation):
        """Returns packing in dictionary format.
        packing algorithm needs topology in DFS order format
        packing = {'name' : [x, y, z, w, h, d], ...}
        
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the created packing.
        @rtype: dictionary
        """
        H = self.__CreateH(representation, self.benchmark.dimensions)
        self.__TreePacking(H)
        packing = {}
        for node in H:
            packing[node[self.__H_NAME]] = [node[self.__H_X],
                                            node[self.__H_Y],
                                            node[self.__H_Z],
                                            node[self.__H_WIDTH],
                                            node[self.__H_HEIGHT],
                                            node[self.__H_DEPTH]]
        return packing

    def get_random_representation(self):
        """Gets a random representation.
        This function gets a random representation out of all possible representations.
        
        @return: Returns a random representation out of all possible representations.
        @rtype: list
        """
        # generate random module permutation
        module_permutation = EC.get_permutation(self.benchmark.blocks,
                                                randrange(self.part_lengths[0]))
        # generate random ternary tree topology
        T, root = TT.get_random_ternary_tree(self.num_blocks)
        topology = TT.transform_panholzer_to_topology(T, root)
        # generate random module_rotations
        module_rotations = EC.get_variation([self.ORIGINAL, self.ROTATED],
                                            self.num_blocks, 
                                            randrange(self.part_lengths[2]))
        # return as t_tree format
        return [module_permutation, topology, module_rotations]

    def packing_to_representation(self, packing):
        """Creates a representation out of a packing.
        
        @param packing: a packing of a current representation
        @type packing: dictionary
        
        @note: Not implemented and/or available!
        """
        print "Not implemented and/or available!"

    def get_operations(self):
        """Returns a dictionary of possible operations onto a abstract solution.
        The higher the globality_factor, the bigger is the solution change
        This factor is given by the authors but it also can be determined
        in test runs evaluation the cost influences caused by applying the
        according function.
        
        @return: Returns a dictionary of possible operations onto a abstract solution.
        structure: {'name' : [function, globality_factor]}
        @rtype: dictionary
        """
        return {'Move node' : [self.move_node_in_topology, 30],
                'Swap two modules' : [self.swap_two_modules, 20],
                'Rotate single block in plane (keep z)': [self.rotate_module, 10]}
        # + evtl. exchange subtrees

    def merge_representations(self, representation_a, representation_b):
        """Merges two representations into all possible children.
        Typically, used for genetic and/or evolutionary algorithms.
        Returns a list with all possible children.
        
        In case of T-Tree all possible combinations of the
        module permutations, topologies second and rotations
        are generated. Rotations are currently completely taken from
        representation a or b, respectively. Thus 2^3 = 8 different children
        are possible.
        
        @param representation_a: a representation list
        @type representation_a: list
        
        @param representation_b: a representation list
        @type representation_b: list
        
        @return: Returns a list with all possible representations created by merging the two given.
        @rtype: list
        """
        result = []
        for n in range(2**3):
            child = []
            for i in range(3):
                if n // 2**(2-i):
                    child.append(representation_a[i])
                else:
                    child.append(representation_b[i])
                n = n % 2**(2-i)
            result.append(child)
        return result

    def rotate_module(self, representation):
        """Rotates a module in a T-Tree.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        rotations = list(representation[2])
        index = randint(0, len(rotations)-1)
        if rotations[index] == self.ORIGINAL:
            rotations[index] = self.ROTATED
        else:
            rotations[index] = self.ORIGINAL
        return representation[:2]+[rotations]

    def swap_two_modules(self, representation):
        """Swaps two modules in a T-Tree.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        module_permutations = list(representation[0])
        pos1 = pos2 = randint(0, len(module_permutations)-1)
        while pos1 == pos2:
            pos2 = randint(0, len(module_permutations)-1)
        module_permutations[pos1], module_permutations[pos2] = module_permutations[pos2], module_permutations[pos1]
        return [module_permutations] + representation[1:]

    def __decrease(self, value, boundary):
        """Decreases value by 1 if not None and if greater than boundary.
        @param value: value to decrease
        @type value: int
        
        @param boundary: value can not get lower than this
        @type boundary: int
        
        @return: Returns value-1 or value, if one statement is violated.
        @rtype: int
        """
        if value == None or value < boundary:
            return value
        else:
            return value - 1 

    def __increase(self, value, boundary):
        """Increases value by 1 if not None and if greater than boundary.
        @param value: value to increase
        @type value: int
        
        @param boundary: value can not get lower than this
        @type boundary: int
        
        @return: Returns value-1 or value, if one statement is violated.
        @rtype: int
        """
        if value == None or value < boundary:
            return value
        else:
            return value + 1 
        
    def move_node_in_topology(self, representation):
        """Moves a node in the topology of a T-Tree.
        @param representation: a representation list
        @type representation: list
        
        @return: Returns the changed representation list.
        @rtype: a representation list
        """
        RETURN_POSITIONS = False
        module_permutations = list(representation[0])
        topology = list(representation[1]) # in DFS order
        # pick node X and cut it out
        pos = randint(0, len(module_permutations)-1)
        initial_pos = pos
        # X is leaf node -> cut it
        if topology[pos][1:] == [None, None, None]:
            module_name = module_permutations.pop(pos)
            parent = topology[pos][0]
            index = topology[parent].index(pos)
            topology[parent] =  topology[parent][0:index] + [None] + topology[parent][index+1:]
            del topology[pos]
            for i in range(len(topology)):
                topology[i] = map(lambda x: self.__decrease(x, pos), topology[i])
        # X is inner node with only one child
        elif topology[pos][1:].count(None) == 2:
            module_name = module_permutations.pop(pos)
            child = filter(lambda x: not x == None, topology[pos][1:])[0]
            parent = topology[pos][0]
            if not parent == None:
                index = topology[parent].index(pos)
                topology[parent] =  topology[parent][0:index] + [child] + topology[parent][index+1:]
            topology[child] = [parent] + topology[child][1:]
            del topology[pos] 
            for i in range(len(topology)):
                topology[i] = map(lambda x: self.__decrease(x, pos), topology[i])
        # X has two or three children
        else:
            module_name = module_permutations.pop(pos)
            children = filter(lambda x: not x == None, topology[pos][1:])
            while len(children) > 0:
                child = children[randrange(len(children))]
                child_name = module_permutations.pop(child-1)
                module_permutations.insert(pos, child_name)
                pos = child
                children = filter(lambda x: not x == None, topology[pos][1:])
            parent = topology[pos][0]
            index = topology[parent].index(pos)
            topology[parent] =  topology[parent][0:index] + [None] + topology[parent][index+1:]
            del topology[pos] 
            for i in range(len(topology)):         
                topology[i] = map(lambda x: self.__decrease(x, pos), topology[i])
        # paste this node somewhere else
        # either inner node with a random branch (n*3 possibilities )
        # or external node (n*2+1 possibilities)
        pos = randint(0, 5*len(module_permutations))
        # insert as external node
        if pos < 2*len(module_permutations)+1:
            counter = -1
            for i in range(len(topology)):
                for j in range(3):
                    if topology[i][j+1] == None:
                        counter += 1
                    if counter == pos:
                        insert_pos = i
                        insert_branch = j
                        break
                if counter == pos:
                    break
            next_left = [topology[insert_pos][1:][i] for i in range(insert_branch)]
            next_left = filter(lambda x: not x == None, next_left)
            if next_left == []:
                new_pos = insert_pos + 1
            else:
                next = next_left[-1]
                while not topology[next][1:] == [None, None, None]:
                    next = filter(lambda x: not x == None, topology[next][1:])[-1]
                new_pos = next + 1
            for i in range(len(topology)):
                topology[i] = map(lambda x: self.__increase(x, new_pos), topology[i])
            topology.insert(new_pos, [insert_pos, None, None, None])
            topology[insert_pos] = topology[insert_pos][0:insert_branch+1] + [new_pos] + topology[insert_pos][insert_branch+2:]
            module_permutations.insert(new_pos, module_name)
            modified_pos = new_pos
        # insert as new internal node (including root)
        else:
            pos -= 2*len(module_permutations)+1
            branch = pos % 3
            pos = pos // 3
            parent = topology[pos][0]
            for i in range(len(topology)):
                topology[i] = map(lambda x: self.__increase(x, pos), topology[i])
            topology.insert(pos, [parent, None, None, None])
            topology[pos] = topology[pos][0:branch+1] + [pos+1] + topology[pos][branch+2:]
            topology[pos+1] = [pos] + topology[pos+1][1:]
            if not parent == None:
                ref_index = topology[parent].index(pos+1)
                topology[parent] = topology[parent][0:ref_index] + [pos] + topology[parent][ref_index+1:]
            module_permutations.insert(pos, module_name)
            modified_pos = pos
        if RETURN_POSITIONS:
            return initial_pos, modified_pos, [module_permutations, topology, representation[2]]
        else:
            return [module_permutations, topology, representation[2]]
        
        