'''configuration file for the GUI

This module contains several parameters for the GUI. They are provided in a "crypted" form using dictionaries and lists.
That seems complicated at the first time. But: Lists and dictionaries can be easily iterated. So the code in the GUI module need not to be programmed
with several controls like if elif else.
For example a algorithm has different parameters, which should all stand together in one group box. The parameters have different values, step sizes etc.
Without a list the GUI must handle every single parameter on its own. With a list, the GUI just reads the correct entry from the list where it is told what to do.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>

@var DATA_STRUCTURE_LIST: Is a dictionary with information about the supported data structures.
The keys of the dictionary are the names of the data structures.

The value of each key is a list which contains:
    - class of the data structure
    - description of the data structure which is displayed in the info box in the GUI
@type DATA_STRUCTURE_LIST: dictionary

@var BENCHMARK_LIST: Is a dictionary with information about the supported benchmarks.
The keys of dictionary are the file names of the benchmarks in the default benchmarks folder.

The value of each key is a list which contains:
    - class of the benchmark parser
    - number of blocks in the benchmark
    - description of the benchmark which is displayed in the info box in the GUI
@type BENCHMARK_LIST: dictionary

@var OPTIMIZATION_LIST: Is a dictionary with information about the supported optimization algorithms.
The keys of the dictionary are the names of the algorithms.

The value of each key is a list which contains:
    - function of the algorithm
    - description of the algorithm which is displayed in the info box in the GUI
@type OPTIMIZATION_LIST: dictionary

@var PARAMETER_LIST: Is a dictionary with the parameters of the algorithms.
The keys of the dictionary are the names of the algorithms.
The values of each key contain a dictionary with the parameter names as keys.

The value of those keys in the sub-dictionary is a list which contains:
    - the data type of the parameter, see DATATYPE_LIST
    - the tooltip which appears after a moment selection the parameter in the GUI
    - values for the range of the parameter, see LIMITFUNCTION_LIST
@type PARAMETER_LIST: dictionary

@var DATATYPE_LIST: Is a dictionary which contains the information about the interface for a data type.
The keys are the names of the data types used for the parameter. The value of each key is a PyQt class.
This class is used as interface between user and GUI.
@type DATATYPE_LIST: dictionary

@var LIMITFUNCTION_LIST: Is a dictionary which contains the functions to adjust the range of the parameters.

The keys of the dictionary are the data types, see DATATYPE_LIST
The values are functions of the PyQt classes. These functions allow a adjustment of
the parameters. The values for the functions are stored in PARAMETER_LIST.
@type LIMITFUNCTION_LIST: dictionary

@var READFUNCTION_LIST: IS a dictionary which contains the read functions for the parameters.

The keys of the dictionary are the data types, see DATATYPE_LIST
The values are functions of the PyQT classes. These functions allow a read of the current setting of the parameters.
@type READFUNCTION_LIST: dictionary

@var HEADER_LIST: Is a list for a table header.

This list is the default header of the planned trials table in the GUI.
@type HEADER_LIST: list

@var EVALUATION_PLOT_LIST: Is a dictionary which contains the supported evaluation plots.
The keys of the dictionary are the names of the evlation plot.

The values of each key contains a list with:
    - function of the evalation plot
    - tooltip for the button in the GUI
@type EVALUATION_PLOT_LIST: dictionary

@var DIAGRAM_PARAMETER_LIST: Is a dictionary which contains the parameter for the evalation plots.
This a nested structure. Every key of the dictionary contains either another dictionary or at the last level a list.
The keys are used in the interface between the GUI and the evaluation functions.
The last level list is referenced by the name of the parameter for the plot.
This list contains:
    - data type of the parameter ,  see DATATYPE_LIST
    - tooltip description of the parameter
    - values for the range functions ,  LIMITFUNCTIONS_LIST

@type DIAGRAM_PARAMETER_LIST: dictionary

'''
############### DATA STRUCTURE LIST ###########################################
from dsc_suite.ds.slicing_tree import SlicingTree
from dsc_suite.ds.sequence_triple import SequenceTriple
from dsc_suite.ds.sequence_quintuple import SequenceQuintuple
from dsc_suite.ds.o_sequence import OSequence
from dsc_suite.ds.t_tree import TTree
from dsc_suite.ds.moving_block_sequence import MBS
from dsc_suite.ds.data_structure import COST_CRITERIA_LIST

""" dictionary of functional implemented data structures
key ... name
value ... [DataStructureClass, 'Information']
"""
#the descripbtion should not go beyond the given text label, split it to more lines otherwise the string does not fit the windows
#                                                                   'Optimizes the floorplan for a benchmark with a Simulated Annealing\n'
DATA_STRUCTURE_LIST = {'3D Slicing Tree'          : [SlicingTree, 'This is the 3D Slicing Tree data structure.'],
                       'Sequence Triple'          : [SequenceTriple, 'This is the Sequence Triple data structure.'],
                       'Sequence Quintuple'       : [SequenceQuintuple, 'This is the Sequence Quintuple data structure.'],
                       'O-Sequence'               : [OSequence, 'This is the O-Sequence data structure.'],
                       'T-Tree'                   : [TTree, 'This is the T-Tree data structure.'],
                       '3D Moving Block Sequence' : [MBS, 'This is the 3D-MBS layout representation']}


############### BENCHMARK LIST ################################################
from dsc_suite.tools.benchmarks import YAL
from dsc_suite.tools.benchmarks import GSRC
from dsc_suite.tools.benchmarks import ACADEMIC

""" dictionary of tested benchmarks
key ... name
value ... [BenchmarkClass, num_blocks, 'Information']
usage: bm = BenchmarkClass(name)
"""
BENCHMARK_LIST = {'apte.yal'          : [YAL, 9, 'apte\n\n number of blocks:\t 9\n benchmarkclass:\t YAL'],
                  'xerox.yal'         : [YAL, 10, 'xerox\n\n number of blocks:\t 9\n benchmarkclass:\t YAL'],
                  'hp.yal'            : [YAL, 11, 'hp\n\n number of blocks:\t 11\n benchmarkclass:\t YAL'],
                  'M198.yal'          : [YAL, 33, 'M198\n\n number of blocks:\t 33\n benchmarkclass:\t YAL'],
                  'ami33.yal'         : [YAL, 33, 'ami33\n\n number of blocks:\t 33\n benchmarkclass:\t YAL'],
                  'ami49.yal'         : [YAL, 49, 'ami49\n\n number of blocks:\t 49\n benchmarkclass:\t YAL'],
                  'playout.yal'       : [YAL, 56, 'playout\n\n number of blocks:\t 56\n benchmarkclass:\t YAL'],
                  'cube.academic'     : [ACADEMIC, 7, 'cube\n\n number of blocks:\t 7\n benchmarkclass:\t ACADEMIC'],
                  'sim_cube.academic' : [ACADEMIC, 4, 'sim_cube\n\n number of blocks:\t 4\n benchmarkclass:\t ACADEMIC'],
                  'layer.academic'    : [ACADEMIC, 10, 'layer\n\n number of blocks:\t 10\n benchmarkclass:\t ACADEMIC'],
                  'apte'              : [GSRC, 9, 'apte\n\n number of blocks:\t 9\n benchmarkclass:\t GSRC'],
                  'xerox'             : [GSRC, 10, 'xerox\n\n number of blocks:\t 10\n benchmarkclass:\t GSRC'],
                  'hp'                : [GSRC, 11, 'hp\n\n number of blocks:\t 11\n benchmarkclass:\t GSRC'],
                  'ami33'             : [GSRC, 33, 'ami33\n\n number of blocks:\t 33\n benchmarkclass:\t GSRC'],
                  'ami49'             : [GSRC, 49, 'ami49\n\n number of blocks:\t 49\n benchmarkclass:\t GSRC'],
                  "n10"               : [GSRC, 10, 'n10\n\n number of blocks:\t 10\n benchmarkclass:\t GSRC'],
                  "n30"               : [GSRC, 30, 'n20\n\n number of blocks:\t 30\n benchmarkclass:\t GSRC'],
                  "n50"               : [GSRC, 50, 'n50\n\n number of blocks:\t 50\n benchmarkclass:\t GSRC'],
                  "n100"              : [GSRC, 100, 'n100\n\n number of blocks:\t 100\n benchmarkclass:\t GSRC'],
                  "n200"              : [GSRC, 200, 'n200\n\n number of blocks:\t 200\n benchmarkclass:\t GSRC'],
                  "n300"              : [GSRC, 300, 'n300\n\n number of blocks:\t 300\n benchmarkclass:\t GSRC']}


############### OPTIMIZATION LIST #############################################
from dsc_suite.opt.monte_carlo_adjusted import generate_monte_carlo_data
from dsc_suite.opt.evolutionary_algorithm import generate_evolution_data
from dsc_suite.opt.simulated_annealing_2 import generate_simulated_annealing_data
from dsc_suite.opt.great_deluge_algorithm import generate_great_deluge_data
from dsc_suite.opt.evaluate_operations import generate_evaluation_operation_data
from dsc_suite.opt.evaluate_operations_mc import generate_evaluation_operation_mc_data

""" dictionary of working optimization approaches 
key ... name
value ... [generation_function, 'Information']
usage: generation_function(parameters**)
Detailed info about the generation_function should be found
in the corresponding docstring.
"""
#the descripbtion should not go beyond 'Simulated Annealing\n' otherwise the string does not fit the windows !!
OPTIMIZATION_LIST = {'Monte Carlo'            : [generate_monte_carlo_data         , 'Evenly distributed sampling of the given solution space.'],
                     'Simulated Annealing'    : [generate_simulated_annealing_data , 'Optimizes the floorplan for a benchmark with a Simulated Annealing\n'
                                                                                     'algorithm. This algorithm accepts worse solutions with a decreasing\n'
                                                                                     'probability. So it is possible to get over a local extreme in the\n'
                                                                                     'cost function.'],
                     'Evolutionary Algorithm' : [generate_evolution_data           , 'Optimizes the floorplan for a benchmark with an Evolutionary Algorithm\n'
                                                                                     'It will generate a start population of random solutions, the parents.\n'
                                                                                     'After that it will use two parents to create children. It uses all\n'
                                                                                     'possbile parent combinations. When the children have been created,\n'
                                                                                     'the next parent generation is picked out. The best solution is always\n'
                                                                                     'amoung them, but also very worse solutions to keep the genetic pool\n'
                                                                                     'in a very wide area.'],
                     'Great Deluge Algorithm' : [generate_great_deluge_data        , 'Optimizes the floorplan for a benchmark with a Great Deluge Algorithm\n'
                                                                                     'The algorithm only accepts solutions which are better then the current\n'
                                                                                     'solution and which are beyond a specific value of the cost function.\n'
                                                                                     'This value is decreased with every solution found by the algorithm.\n'
                                                                                     'This algorithm works very fast, but can get stuck in a local extreme.'],
                     'EvalOperation Algorithm': [generate_evaluation_operation_data ,'This algorithm generates data to evaluate the data structure operations\n'
                                                                                     'to change a given solution. It will create several basic solutions.\n'
                                                                                     'A local search around these solutions will be done.']}
#pass for documentaion with epydoc
pass
############### PARAMETER LIST OF ALGORITHMS ##################################
""" this dictionary contains the different parameters of the algorithms
PARAMETER_LIST[name_of_algorithm] returns a dictionary:
keys : parameter name used in algorithm
value : list with detailed information about parameter

index and refered information:
0 : data type -> see DATATYPE_LIST
1 : description for tool tip
2 : list with limits ,  MUST be correlated with functions in LIMITFUNCTION_LIST !

that means: the first object in the list is used as a parameter for the first function for that datatype
"""

PARAMETER_LIST = {"Monte Carlo"            : {"samples"        : ["int", "Created solutions to sample the solution space.", [1000, 10000000, 1000, 10000]]},
                  
                  "Simulated Annealing"    : {"Tsamples"       : ["int", "Created solutions in a temperature step to improve cost function.", [50, 2000, 5, 100]],
                                              "Tsteps"         : ["int", "The steps between Tstart and Tend.", [10, 100, 5, 25]],
                                              "Tstart"         : ["int", "The start temperature for the algorithm.", [1300, 5000, 100, 1300]],
                                              "Tend"           : ["int", "The end temperature for the algorithm.", [50, 350, 50, 100]],
                                              "correct_factor" : ["float"  , "The correct factor for the e-function in the algorithm.", [6, 0.000001, 0.1, 0.000005, 0.0001]]},
                                           
                  "Evolutionary Algorithm" : {"parents"     : ["int", "The number of parent individuals per generation.", [2, 50, 2, 8]],
                                              "children"    : ["int", "The number of created children per recombination.", [1, 20, 1, 4]],
                                              "generations" : ["int", "The number of created generations.", [5, 100, 5, 50]]},
                                              
                  "Great Deluge Algorithm" : {"vapor_samples": ["int", "Created solutions to improve the cost function before the algorithm is left.", [500, 100000, 100, 3000]],
                                              "vapor_value"  : ["float", "Difference between two critical values.", [3, 0.001, 0.02, 0.001, 0.005]]},
                  
                  "EvalOperation Algorithm": {"basic_solutions"  : ["int", "The number of basic solutions.", [1, 100000, 100, 500]],
                                              "derivations"      : ["int", "The number of derivation depth from basic solution.", [0, 3, 1, 0]],
                                              "derived_solutions": ["int", "The number of created solutions in each derivation", [1, 500, 50, 100]],
                                              "cancel_after_trial": ["int", "The number of iterations after the algorithm stops if no new solution is found with the operation."
                                                                     "If value is '0' the algorithm will run through all iterations." , [0, 10, 1, 2]]}}
#pass for documentaion with epydoc
pass
############### DATATYPE LIST #############################################
""" this dictionary contains the different data types which are used in the algorithms and combines them with the Qt-Framework
keys  : used data type
value : used Qt Widget

for example:
to set a integer based parameter a QtSpinBox is used

Hint: Linking to PyQt without import ! ,  string must be run with eval(string) in main program
"""
DATATYPE_LIST = {"int"   : "QtGui.QSpinBox()",
                 "float" : "QtGui.QDoubleSpinBox()",
                 "str"   : "QtGui.QLineEdit()"}

#pass for documentaion with epydoc
pass
############### LIMITFUNCTION LIST #############################################
""" this dictionary contains the different functions to set the limits of the parameters
keys  : used data type
value : list of functions for setting the limits

MUST be correlated to the limits in PARAMETER_LIST

for example:
to set a integer based parameter a QtSpinBox is used 
"""
LIMITFUNCTION_LIST = {"int"   : ["setMinimum",
                                 "setMaximum",
                                 "setSingleStep",
                                 "setValue"],
                                   
                      "float" : ["setDecimals",
                                 "setMinimum",
                                 "setMaximum",
                                 "setSingleStep",
                                 "setValue"],
                      "str"   : ["setText"]}

#pass for documentaion with epydoc
pass
############### READFUNCTION LIST #############################################
""" this dictionary contains the different functions to read the parameters
keys  : used data type ,  used as cast operator
value : list of functions to read the parameter


for example:
to read a integer based parameter from a QSpinBox is used function value 
"""
READFUNCTION_LIST = {"int"   : "value",
                     "float" : "value",
                     "str"   : "text"}

############### HEADER LIST ##################################################
""" this list contains the names of parameters which are used from every algorithm
cost_criteria is replaced by the cost criteria key list in datastructure module
"""

HEADER_LIST = ["#", "trial name" , "datastructure", "benchmark", "algorithm", "z-Dimension"]
for key in COST_CRITERIA_LIST.keys():
    HEADER_LIST += [key, "weight factor " + key, "norm value " + key]
del key

############### PLOT METHODS LIST #############################################
from dsc_suite.analyses.cost_distribution import plot_cost_distribution
from dsc_suite.analyses.cost_distribution_2D import plot_2d_cost_distribution
from dsc_suite.analyses.cost_development import plot_costdata_development
from dsc_suite.analyses.scatter_plot_2D import plot_2D_scatter_plot
from dsc_suite.analyses.histogram import plot_histogram
from dsc_suite.analyses.scatter_plot_op import plot_scatter_plot_operation
from dsc_suite.analyses.histogram_op import plot_histogram_operation

""" dictionary of evaluation methods which generate a plot
key ... name
value ... [generation_function, 'Information']
usage: generation_function(parameters**)
Detailed info about the generation_function should be found
in the corresponding docstring.
"""
EVALUATION_PLOT_LIST = {'Cost Distribution' : [plot_cost_distribution,
                                               'Plots a cost distribution.'],
                        '2D Cost Distribution' : [plot_2d_cost_distribution,
                                                  'Plots two-dimensional cost distribution.'],
                        'Cost development': [plot_costdata_development,
                                                 'Plots cost data over their index in data file'],
                        'Scatter plot' : [plot_2D_scatter_plot,
                                             'Plots a scatter plot of two cost criteria.'],
                        'Histogram' : [plot_histogram, 'Plots the envelope of the histogram of the data.'],
                        
                        'Scatter Plot Operation' : [plot_scatter_plot_operation, ''],
                        
                        'Histogram Operation' : [plot_histogram_operation, '']}
#pass for documentaion with epydoc
pass
"""dictionary of diagram options which can be adjusted
keys ... will create tabs (as first level) and group boxes all sub-levels
value ... dictionary ,  next level of group box, list ,  information about parameter
usage: similar to parameter list
The DSC constructs a dialog with a special layout. The first level keys create tabs.
If the value referenced by the key is a dictionary, all sub-level keys will create a group box and include all next sub-level keys.
If the value referenced by the key is a list, the key is a parameter and the DSC will create a PyQt item for adjustment (see DATATYPE_LIST).
The other items in the list are tooltip and limits (if needed) ,  see limit function list).
"""
DIAGRAM_PARAMETER_LIST = {'labels' : {'title'            : ["str" , "The title of the diagram.", ""],
                                      'legend-position'  : ["str" , "The position of the legend. Possible options:\nbest\nupper right\nupper left\nlower left\nlower right\nright\ncenter left\ncenter right\nlower center\nupper center\ncenter", "best"],
                                      'legend-title'     : ["str" , "The title of the legend.", ""],
                                      'legend-parameter' : ["str" , "The parameter which is different in the selected files. Only one supported!", ""],
                                      'x-label'          : ["str" , "The label for the x-axe.", ""],
                                      'y-label'          : ["str" , "The label for the y-axe.", ""]},
                                      
                          'sizes' : {'dimensions' : {'length' : ["int" , "The length of the diagram.", 1, 30, 1, 10],
                                                     'width'  : ["int" , "The width of the diagram.", 1, 30, 1, 8]},
                                     'margins'    : {'left'   : ["float" , "The left white margin in percent", 2, 0.05, 0.3, 0.05, 0.1],
                                                     'right'  : ["float" , "The right white margin in percent", 2, 0.7, 1.0, 0.05, 0.9],
                                                     'bottom' : ["float" , "The bottom white margin in percent", 2, 0.05, 0.3, 0.05, 0.1],
                                                     'top'    : ["float" , "The top white margin in percent", 2, 0.7, 1.0, 0.05, 0.9]}},
                          'axes' : {'x-axe'       : {'xmin'   : ["str" , "The minimum x-value. Auto = 0", "auto"],
                                                     'xmax'   : ["str" , "The maximum x-value. Auto = searches maximum", "auto"], },
                                    'y-axe'       : {'ymin'   : ["str" , "The minimum y-value. Auto = 0", "auto"],
                                                     'ymax'   : ["str" , "The maximum y-value. Auto = searches maximum", "auto"]}}}
