'''implements a XMLParser for the GUI

This module provides the class XMLParser especially designed for the use with the GUI.
You can create new XML files, write new data into an existing file and you can read information from a file.
There is already a search function implemented which gives back specific information.
This function can be used as a template for a more general search.

The communication interface is based on dictionaries.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''
import xml.dom.minidom as DOM
import os
#from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA as FFDATA

class XMLParser(object):
    """Provides a XML File Parser for the GUI.
    This class is the parser for the XML files. It is based on dictionary interface to the GUI.
    Every key in a dictionary is used as a tag name in the XML file. The value is either the tag content or (when a dictionary) the next hierarchical level of the XML file.
    For better understanding please refer to the source code.
    """
    def __init__(self, source, mode):
        """Creates a new XMLParser instance.
        This is the constructor for the class. It parses the XML file given by path. If the XML file does not exist, it will be created.
        The mode option is used for the tag name of the top level entries.
        For example mode = 'datafile', the XML structure is:
            - <datafiles>
                - <datafile>
                    - ...
                - </datafile>
            - </datafiles>    
        @param source: Is the relative path to the XML file from the default folder for data.
        @type source: string

        @param mode: Is the top-level tag.
        Supported in this version:
            - 'datafile'
            - 'plot'
        @type mode: string
        """
        self.__source = source
        self.__mode = mode
        if not os.path.exists(self.__source):
            self.createNewXML(self.__source, mode+"s")
        self.__tree = DOM.parse(self.__source)
            
    def nextTopLevelTag(self, index):
        """Returns the top level tag at position index with all sub-tags.
        This function splits the content of the XML into all top level tags with their sub-tags.
        It returns a dictionary with the structure of the content of the XML file.
        All tag names will appear as keys in the dictionary.
        The value of a key is either a dictionary or a string.
        A dictionary is used for the next hierarchical level of the XML file.
        The string contains the text value inside the tag.
        For example:
            - <top>
                - <tag1>
                    - <tag2>value</tag2>
                - </tag1>
            - </top>
        Will return:
            - {'top' : { 'tag1' : { 'tag2' : 'value' }}}
        Hint: Use with countTopLevelItems() to avoid index out of range exception!
        
        @param index: Is the index of the item.
        @type index: int
        
        @return: Returns the top level tag at position index with all sub-tags.
        @rtype: dictionary
        """
        root = self.__tree.firstChild
        #get toplevelitem name
        toplevelname = str(self.__findTopLevelItemName(self.__tree.firstChild))
        #get all toplevelitems
        olddata = self.__tree.getElementsByTagName(toplevelname)
        #get toplevelitem with index
        datafile = olddata[index]
        #create dictionary for all sublevel tags and values
        dict = self.__getTag(datafile, root)
        return dict[toplevelname]
        
    def countTopLevelTags(self):
        """Counts the top level items in a XML file.
        This function counts the top level items in a XML file. The mode option in the constructor must be set correct!
        
        @return: Returns number of top level items in the XML file.
        @rtype: int
        """
        toplevelname = str(self.__findTopLevelItemName(self.__tree.firstChild))
        return len(self.__tree.getElementsByTagName(toplevelname))
    
    def writeNewData(self, dict):
        """Writes the data in dict into a XML file.
        This function writes the content of the dictionary given by dict into the XML file.
        The information will be appended at the end of the file.
        The top level item must not be in the dictionary. The mode option from the contrucutor is used to create it.
        Example:
        structure of dict:
            - { 'key1' : {'key2' : 'value2'}, 'key3' : 'value3' }
        results in XML structure:
            - <root>
                - <mode>
                    - <key1>
                        - <key2>value2</key2>
                    - </key1>
                - </mode>
            - </root>
        @param dict: Information to write into the XML file.
        @type dict: dictionary
        """
        #read old data from tree
        try:
            toplevelname = str(self.__findTopLevelItemName(self.__tree.firstChild))
            rootname = str(self.__tree.firstChild.tagName)
            olddata = self.__tree.getElementsByTagName(toplevelname)
        except:
            toplevelname = None
        
        if toplevelname == "None":
            toplevelname = self.__mode
            rootname = self.__mode+"s"
        if toplevelname != self.__mode:
            raise AttributeError
        #getting root an old data
        
        #create new XML tree
        doc = DOM.Document()
        root = DOM.Element(rootname)
        #add old data to new tree - recurrent
        for old_top in olddata:
            self.__newFromOldNode(old_top, root)         
        
        #adding new data - recurrent
        for index in dict.keys():
            newtop = DOM.Element(toplevelname)
            self.__newNode(dict[index], newtop)
            root.appendChild(newtop)
        
        #print to a pretty XML with same path --> overriding old file
        doc.appendChild(root) 
        fobj = open(self.__source, "w")
        print >> fobj, doc.toprettyxml()
        fobj.close()
    
    def createNewXML(self, path, root):
        """Creates a new XML file.
        This function is used to create a new XML file at the position path. The root instance name is root.
        
        @param path: The path where the XML is created.
        @type path: string
        
        @param root: The name of the root instance.
        @type root: string
        """
        #create document instance
        doc = DOM.Document()
        #create root
        root = DOM.Element(root)
        doc.appendChild(root)
        #write XML file
        fobj = open(path, "w")
        print >> fobj, doc.toprettyxml()
        fobj.close()
    
    def getRootName(self):
        """Returns the name of the root instance in a XMl file.
        @return: Returns the name of the root instance in a XMl file.
        @rtype: string
        """
        return str(self.__tree.firstChild.tagName)
    def getCharacteristic(self, data, algorithm, bench, cost_criteria, characteristics):
        """Extracts specific information from a XML file.
        This function is used to extract specific information from the XML file.
        It can be used as a template for a more abstract function.
        This function was created to get the mean values from the mean_values.xml without opening it in the GUI.
        So this function is only useful for getting characteristic information.
        It will only work for the mean_values.xml, because there is not more than one entry of the searched combination
        
        Exmaple of return:
            - { 'criteria 1' : { 'char1' : 'value' , 'char2' : 'value'}, 'criteria 2' : { 'char1' : 'value', 'char2' : 'value'}}
            
        @param data: The searched data structure.
        @type data: string
        
        @param algorithm: The searched algorithm.
        @type algorithm: string
        
        @param bench: The searched benchmark.
        @type bench: string
        
        @param cost_criteria: The searched criteria.
        @type cost_criteria: list-of-strings
        
        @param characteristics: The searched characteristics.
        @type characteristics: list-of-strings
        
        @return: Dictionary with the information, empty if not found.
        @rtype: dict
        
        
        """
        #getting XML tree
        root = self.__tree.firstChild
        #get main entries
        items = root.getElementsByTagName("datafile")
        #copy list
        new_items = items[:]
        #prepare search lists
        tag_list = ["datastructure", "algorithm", "benchmark"]
        value_list = [data, algorithm, bench]
        #search in the XML for the specific entry
        #first step find all items with datastructure --> new search pool
        #second step find in new pool all items with algorithm --> new search pool
        #third step find in new pool all items with benchmark
        for i in range(0,len(tag_list), 1):
            items = new_items[:]
            new_items = []
            for item in items:
                list = item.getElementsByTagName(tag_list[i])
                for node in list:
                    child = node.firstChild
                    text = str(child.data.strip())
                    if text == value_list[i]:
                        new_items.append(item)
        #now find in the entry the criteria given by list
        #in the criteria level find characteristics given by list
        #create dictionary for return            
        items = new_items[:]
        chars = {}
        for string in cost_criteria:
            for item in items:
                list = item.getElementsByTagName(string)
                for criteria in list:
                    chars.update({string: {}})
                    for char in characteristics:
                        tag = criteria.getElementsByTagName(char)[0]
                        text = tag.firstChild.data.strip()
                        chars[string].update({char:text})
        return chars
    
    def __findTopLevelItemName(self, node):
        """Used to find the top level name.
        Finds the top level name in a XML file.
        
        @param node: Root instance node.
        @type node: DOM.node
        
        @return: Name of the top level items.
        @rtype: string
        """
        for child in node.childNodes:
            if child.nodeType == DOM.Node.ELEMENT_NODE:
                if child.tagName == "datafile" or child.tagName == "plot":
                    return child.tagName                    
            else:
                continue
            
    def __newNode(self, todo, parent):
        """Used to create a new node.
        Builds the DOM structure for writing into a XML file.
        Recursive function.
        
        @param todo: Dictionary with information to write.
        @type todo: dictionary
        
        @param parent: Parent node of the DOM structure.
        @type parent: DOM.node
        """
        #iterate todo
        for key in todo.keys():
            #new Element node with sub-tags ...
            if type(todo[key]) == dict:
                    newchild = DOM.Element(key)
                    self.__newNode(todo[key], newchild)
            #...or 'real' information tag with value        
            else:
                newchild = DOM.Element(key)
                newtext = DOM.Text()
                newtext.data = todo[key]
                newchild.appendChild(newtext)
            #do not forget the appending to the parent!
            parent.appendChild(newchild)
            
    def __newFromOldNode(self, node, parent):
        """Used for 'old' information in the XML file.
        This function appends the old information to the new DOM structure while writing into the XML file.
        Otherwise this information would be lost!
        Recursive function.
        
        @param node: Node in the old DOM structure.
        @type node: DOM.node
        
        @param parent: Parent node in the new DOM structure
        @type parent: DOM.node
        """
        #differ between element node...  
        if node.nodeType == DOM.Node.ELEMENT_NODE:
            newnode = DOM.Element(node.tagName)
            #go deeper into structure
            if node.hasChildNodes:
                for child in node.childNodes:
                    self.__newFromOldNode(child, newnode)
            parent.appendChild(newnode)
        #... and text node with information, exclude white spaces, otherwise new XML will not look 'pretty printed'    
        if node.nodeType == DOM.Node.TEXT_NODE and node.nextSibling == None and node.previousSibling == None:
            newtext = DOM.Text()
            newtext.data = node.data.strip()
            parent.appendChild(newtext)
    
    def __getTag(self, node, parent):
        """Reads the information from the XML file.
        This function is the 'real' reading function. It walks down the structure of a top level item and extracts all information.
        It also builds the dictionary which is returned to the GUI.
        Recursive function.
        
        @param node: Node in the DOM structure
        @type node: DOM.node
        
        @param parent: Parent of node in the DOM structure.
        @type parent: DOM.node
        
        @return: Returns a dictionary with structure and information found in XML file.
        @rtype: dictionary
        """
        #deside wether Element Node...
        if node.nodeType == DOM.Node.ELEMENT_NODE:
            dict = {}
            if node.hasChildNodes:
                #for each child node call function again
                #--> until all information is in dict
                for child in node.childNodes:
                    newdict = self.__getTag(child, node)
                    if node.tagName not in newdict.keys():
                        try:
                            dict[node.tagName].update(newdict)
                        except KeyError:
                            dict[node.tagName] = {}
                            dict[node.tagName].update(newdict)
                    else:
                        dict = newdict
                return dict
        #...or TextNode with interesting content --> skipping whitespace tags                        
        if node.nodeType == DOM.Node.TEXT_NODE and node.nextSibling == None and node.previousSibling == None:
            #creating content of dict
            key = str(parent.tagName)
            text = node.data.strip()
            text = str(text)
            return {key: text}
        else:
            return {}