'''Implements a algorithm to evaluate the operations of a data structure

This module implements a algorithm which can be used to evaluate the operations of a data structure. The operations take a given solution and return a changed one.
This algorithm is used to create a amount of base solutions. These solutions will be changed multiple times by the operations. Each time the base solution is the same!
That means, it is a localized search around different spots in the solution area. The created data files contain the basic solutions and the created ones. An interation (ITER)
file will be created with information abount basic solution number, number of created solution and number of operation index.

@attention: The data file structure is different to the other algorithms implemented in this module. Nether the less, this algorithm can be called in the same way like the others. The evaluation function for this set of data must be special and the data will not work with other evaluation tools.


@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''
import os, numpy
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA
from time import time
from multiprocessing.process import Process
from multiprocessing.queues import Queue
from multiprocessing import cpu_count

class Worker(Process):
    
    def __init__(self, solution, operation, fcost, amount_derived, index_code, lenght_cost_data, transfer):
        Process.__init__(self)
        self.__solution = solution
        self.__operation = operation
        self.__fcost = fcost
        self.__amount_derived = amount_derived
        self.__index_code = index_code
        self.__lenght_cost = lenght_cost_data
        self.__transfer = transfer
    
    def run(self):
        new_solutions = []
        cost_data = []
        for k in range(0,self.__lenght_cost,1):
            cost_data.append([])
            
        for i in range(0,self.__amount_derived,1):
            self.__index_code += "-"+str(i)
            new_solution = self.__operation(self.__solution)
            k = 0
            while new_solution in new_solutions or k <= self.__amount_derived:
                new_solution = self.__operation(self.__solution)
                k += 1
            new_solutions.append(new_solution)
            if k == self.__amount_derived:
                print "Overflow on possible solutions"
            ncosts = self.__fcost(new_solution)
            for ncost in ncosts:
                    index = ncosts.index(ncost)
                    cost_data[index].append(ncost)
            cost_data[len(cost_data)-1].append(self.__index_code)
            self.__transfer.put(cost_data)


def generate_evaluation_operation_mc_data(functions, parameters, file_info, time_check = False):
    '''
    
    @param functions: Contains functions for the calculation.
    
    essential keys:
        - "randomSolution" : Returns a random representation.
        - "getOperations" :  Returns a list with functions to change a solution.
        - "costEvaluation" : Returns the calculated costs for the given representation.
    @type functions: dictionary    
    
    @param parameters: Contains algorithm parameters.
    
    essential keys:
        - "basic_solutions" : number of basic solutions used to sample solution space
        - "derived_solutions" : number of solutions which will be derived from a basic solution
    @type parameters: dictionary           
    
    @param file_info: Contains information about the filename of the created file.
    
    essential keys:
        - "trial_name" : name of trial given in GUI
        - "file_name_list" : additional information included in file name (can be empty string)
    @type file_info: dictionary
    
    @param time_check: Sets the function to estimate runtime or full calculation.
        - True: estimates runtime with reduced samples
        - False: full calculation with saving data to files
    @type time_check: boolean
    
    @return: Returns either a list with filenames (time_check == False) or a float with the estimated runtime (time_check == True).
    @rtype: list, float
    '''      
    #psyco should speed python a little bit
    start = time()
    import psyco
    psyco.full()
    #----------------------INITIALIZATION----------------------------
    basics  = []
    cost_data = []
    operations = functions["getOperations"]()
    if time_check:
        amount_derived = 50
        amount_basics = 1
    else:
        amount_derived = parameters["derived_solutions"]
        amount_basics = parameters["basic_solutions"]
        
    for i in range(0,amount_basics,1):
        basics.append(functions["randomSolution"]())
    costs = functions["costEvaluation"](basics[0])
    for cost in costs:
        cost_data.append([])
    cost_data.append([])
    
    for solution in basics:
        costs = functions["costEvaluation"](solution)
        for cost in costs:
            index = costs.index(cost)
            cost_data[index].append(cost)
        index_code = "b-"+str(basics.index(solution))
        cost_data[len(cost_data)-1].append(index_code)
             
    #----------------------MAIN ALGORITHM PART--------------------
    #first loop through possible operations
    print "Start"
    for operation in operations:
        print str(operations.index(operation))+" "+str(operation)
        #loop through basic solutions
        for i in range(0,len(basics),cpu_count()/2):
            print i
            workers = []
            queue = Queue()
            for k in range(0,cpu_count()/2,1):
                try:
                    solution = basics[i+k]
                    fcost = functions["costEvaluation"]
                    print k
                    index_code = "d-"+str(operations.index(operation))+"-"+str(basics.index(solution))
                    lenght_cost_data = len(cost_data)
                    transfer = queue
                    
                    workers.append(Worker(solution, operation, fcost, amount_derived, index_code, lenght_cost_data, transfer))
                except:
                    print "Exception"
                    continue
            print workers
            for worker in workers:
                worker.start()
            for worker in workers:
                worker.join()
            while queue.qsize() > 0:
                costs = queue.get()
                for cost in costs:
                    cost_data[costs.index(cost)] += cost
         
    #for j in range(0,len(cost_data[0]),1):
    #    print str(cost_data[0][j])+"\t"+str(cost_data[1][j])+"\t"+str(cost_data[2][j])+"\t"+str(cost_data[3][j])     
    print "End"
    #stop psyco
    end = time()        
    psyco.stop()
    #---------------------TIME ESTIMATION----------------------------
    #estimate runtime
    runtime = end-start
    #return runtime      
    if time_check:
        return runtime*parameters["basic_solutions"]*parameters["derived_solutions"]/(amount_basics*amount_derived)
    #------------SAVING ROUTINE--------------------------------------
    #save files
    #file location
    trial_name = file_info["trial_name"]
    file_name_list = file_info["file_name_list"]
    path = trial_name + '/Evaluate Operations/'
    if not os.path.exists(os.path.join(FOLDER_FOR_DATA, path)):
        os.makedirs(os.path.join(FOLDER_FOR_DATA, path))
    #filename list for return
    filename_list = []
    for cost in cost_data:
        file_name_begin = file_name_list[cost_data.index(cost)]
        i = 1
        filename = file_name_begin + ' %02i B%i D%i.npy' % (i, parameters["basic_solutions"], parameters["derived_solutions"])
        while os.path.exists(os.path.join(FOLDER_FOR_DATA, path, filename)):
            i += 1
            filename = file_name_begin + ' %02i B%i D%i.npy' % (i, parameters["basic_solutions"], parameters["derived_solutions"])
        filename_list.append(os.path.join(path, filename))
        if not time_check:
            numpy.save(os.path.join(FOLDER_FOR_DATA, path, filename), numpy.array(cost))
    #return list of filenames    
    return filename_list
