'''implements a Great Deluge optimization concept

This module implements a Great Deluge optimization concept. Its function can be called either from console or from GUI(recommended).
It saves the solution files in standard folder for data.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''
import random, numpy, os
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA
from time import strftime, time

def generate_great_deluge_data(functions, parameters, file_info, time_check = False):
    '''Optimizes the given floorplan with a Great Deluge algorithm concept and saves it progress in data files.
    
    @param functions: Contains functions for the calculation.
    
    essential keys:
        - "randomSolution" : Returns a random representation.
        - "changeSolution" : Returns a function to change the given representation.
        - "costEvaluation" : Returns the calculated costs for the given representation.
    @type functions: dictionary    
    
    @param parameters: Contains algorithm parameters.
    
    essential keys:
        - "vapor_samples" : number of tried solutions for one critical value
        - "vapor_value" : reduce critical value with every better solution
    @type parameters: dictionary           
    
    @param file_info: Contains information about the filename of the created file.
    
    essential keys:
        - "trial_name" : name of trial given in GUI
        - "file_name_list" : additional information included in file name (can be empty string)
    @type file_info: dictionary
    
    @param time_check: Sets the function to estimate runtime or full calculation.
        - True: estimates runtime with reduced samples
        - False: full calculation with saving data to files
    @type time_check: boolean
    
    @return: Returns either a list with filenames (time_check == False) or a float with the estimated runtime (time_check == True).
    @rtype: list, float
    '''
    #psyco should speed python a little bit
    import psyco
    psyco.full()    
    #time estimation not yet working for great deluge
    runtime = 0.0
       
    if time_check:
        return runtime              
    #-----------MAIN OPTIMIZATION PART----------------
    #init values
    cycle = 0
    critical_value = 1.0
    #force optimization
    while( cycle == 0):
        #generate start solution
        i = 0
        solution = functions["randomSolution"]()
        costs = functions["costEvaluation"](solution)
        cost_data = []
        for cost in costs:
            cost_data.append([])
        cost_data.append([])
        tabu_list = []
        #only until maximum value is reached    
        while (i < parameters["vapor_samples"]):
            #create new solution
            new_solution = functions["changeSolution"]()(solution)
            if new_solution not in tabu_list:
                tabu_list.append(new_solution)
            else:
                i +=1
                continue
            #evaluate solutions
            old_costs = functions["costEvaluation"](solution)
            sum_old = old_costs[len(old_costs)-1]
            new_costs = functions["costEvaluation"](new_solution)
            sum_new = new_costs[len(new_costs)-1]
            #accept solution ?
            if sum_new < critical_value and sum_new < sum_old:
                #yes -- new solution is start solution
                solution = new_solution
                #save new solution
                for cost in new_costs:
                    cost_data[new_costs.index(cost)].append(cost)
                cost_data[len(cost_data)-1].append(cycle)
                #reset abort condition
                i = 0
                #lower water surface
                critical_value -= parameters["vapor_value"]
                #raise cycle value
                cycle += 1
            else:
                #no -- raise abort value
                i += 1
    
    #------------SAVING ROUTINE---------------------------
    #not to forget to stop speeding
    psyco.stop()
    #create new files
    trial_name = file_info["trial_name"]
    file_name_list = file_info["file_name_list"]
    path = trial_name + '/Great Deluge Algorithm/'
    if not os.path.exists(os.path.join(FOLDER_FOR_DATA, path)):
        os.makedirs(os.path.join(FOLDER_FOR_DATA, path))
    filename_list = []
    for list in cost_data:
        file_name_begin = file_name_list[cost_data.index(list)]
        i = 1
        filename = file_name_begin + ' %02i.npy' % (i)
        while os.path.exists(os.path.join(FOLDER_FOR_DATA, path, filename)):
            i += 1
            filename = file_name_begin + ' %02i.npy' % (i)
        filename_list.append(os.path.join(path, filename))
        if not time_check:
            numpy.save(os.path.join(FOLDER_FOR_DATA, path, filename), numpy.float64(list))
    #return list of created files    
    return filename_list