'''implements a Monte Carlo search of the solution space

Generating solutions randomly distributed over the solution space
without any dependencies between them.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>


'''

import numpy, os
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA
from time import time


#TODO: What is with future parameters that are not considered now. How do they find a way into the gui.
#TODO: Multi-threading support. Splitting of samples into sub-tasks. Combination-Routines required.
#TODO: required time estimation
def generate_monte_carlo_data(functions, parameters, file_info,  time_check = False):
    """ Samples solution space randomly, evenly distributed and without dependencies.
    
    
    @param functions: Contains functions for the calculation.
    
    essential keys:
        - "randomSolution" : Returns a random representation.
        - "costEvaluation" : Returns the calculated costs for the given representation.
    @type functions: dictionary    
    
    @param parameters: Contains algorithm parameters.
    
    essential keys:
        - "samples" : number of generated random representations
    @type parameters: dictionary           
    
    @param file_info: Contains information about the filename of the created file.
    
    essential keys:
        - "trial_name" : name of trial given in GUI
        - "file_name_list" : additional information included in file name (can be empty string)
    @type file_info: dictionary
    
    @param time_check: Sets the function to estimate runtime or full calculation.
        - True: estimates runtime with reduced samples
        - False: full calculation with saving data to files
    @type time_check: boolean
    
    @return: Returns either a list with filenames (time_check == False) or a float with the estimated runtime (time_check == True).
    @rtype: list, float
    """
    import psyco #@UnresolvedImport
    psyco.full()
    start_time = time()
    #create data list for saving
    representation = functions["randomSolution"]()
    costs = functions["costEvaluation"](representation)
    cost_data = []
    for cost in costs:
        cost_data.append([])
    #reduce sample for runtime estimation
    if time_check:
        samples = parameters["samples"]
        parameters["samples"] = 10
    #start random generation
    i = 0    
    while i < parameters["samples"]:
        i += 1
        #generate solution
        representation = functions["randomSolution"]()
        #calculate cost criteria
        costs = functions["costEvaluation"](representation)
        #adding cost to saving list
        for cost in costs:
            cost_data[costs.index(cost)].append(cost)
    #save data to files
    trial_name = file_info["trial_name"]
    file_name_list = file_info["file_name_list"]
    path = trial_name + '/Monte Carlo/'
    if not os.path.exists(os.path.join(FOLDER_FOR_DATA, path)):
        os.makedirs(os.path.join(FOLDER_FOR_DATA, path))
    filename_list = []
    for list in cost_data:
        file_name_begin = file_name_list[cost_data.index(list)]
        i = 1
        filename = file_name_begin + ' %02i S%i.npy' % (i, parameters["samples"])
        while os.path.exists(os.path.join(FOLDER_FOR_DATA, path, filename)):
            i += 1
            filename = file_name_begin + ' %02i S%i.npy' % (i, parameters["samples"])
        filename_list.append(os.path.join(path, filename))
        if not time_check:
            numpy.save(os.path.join(FOLDER_FOR_DATA, path, filename), numpy.float64(list))
    end_time = time()
    psyco.stop()
    #return either runtime or filename list
    if time_check:
        runtime = (end_time - start_time) / parameters["samples"] * samples
        parameters["samples"] = samples
        return runtime
    else:
        return filename_list
