'''implements a Simulated Annealing optimization concept

This module implements a Simulated Annealing optimization concept. Its function can be called either from console or from GUI(recommended).
It saves the solution files in standard folder for data.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
'''
import math, numpy, random, os
from dsc_suite.tools.toolpaths import FOLDER_FOR_DATA
from time import strftime, time

def generate_simulated_annealing_data(functions, parameters, file_info, time_check = False):
    '''Optimizes the given floorplan with a Simulated Annealing algorithm concept and saves it progress in data files.
    
          
    @param functions: Contains functions for the calculation.
    
    essential keys:
        - "randomSolution" : Returns a random representation.
        - "changeSolution" : Returns a function to change the given
                             representation
        - "costEvaluation" : Returns the calculated costs for the given representation.
    @type functions: dictionary    
    
    @param parameters: Contains algorithm parameters.
    
    essential keys:
        - "Tstart" : initial temperature of algorithm 
        - "Tend" : temperature to reach with annealing
        - "Tsteps" : steps between start and end temperature
        - "Tsamples" : number of generated solutions each step
        - "correct_factor" : factor to adjust the acceptance probability
    @type parameters: dictionary           
    
    @param file_info: Contains information about the filename of the created file.
    
    essential keys:
        - "trial_name" : name of trial given in GUI
        - "file_name_list" : additional information included in file name (can be empty string)
    @type file_info: dictionary
    
    @param time_check: Sets the function to estimate runtime or full calculation.
        - True: estimates runtime with reduced samples
        - False: full calculation with saving data to files
    @type time_check: boolean
    
    @return: Returns either a list with filenames (time_check == False) or a float with the estimated runtime (time_check == True).
    @rtype: list, float
    '''
    #psyco should speed python a litte bit
    import psyco
    psyco.full()
    if time_check:
        tsamples = parameters["Tsamples"]
        parameters["Tsamples"] = 10
        tsteps = parameters["Tsteps"]
        parameters["Tsteps"] = 3
    #calculate temperature lowering coefficient
    temp = parameters["Tstart"]
    n = 1./(parameters["Tsteps"]-1.)
    alpha = (float(parameters["Tend"])/float(parameters["Tstart"]))**n
    
    #calculates the correct_factor
    #only 1% of the slightly worse solutions should be accepted at the end of the algorithm
    #means a difference of parameters["correct_factor"] will not be accecpted at the end
    #Attention ! correlates with the beginning temperature as well --> could lead to a very low acceptance threshold 
    correct_factor = float(((-1)*math.log(0.1)*float(parameters["Tend"]))/parameters["correct_factor"])
    
    solution = functions["randomSolution"]()
    costs = functions["costEvaluation"](solution)
    
    cost_data = []
    for cost in costs:
        cost_data.append([])
    cost_data.append([])
    
    #------------------------------MAIN OPIMIZATION PART------------------------------
    #until temperature is at Tend
    while temp >= parameters["Tend"]:
        #reset tabu list
        tabu_list = []
        istart = time()
        #just for intern console use
        number_of_acepted = 0
        number_of_impr = 0
        #until maximum new solutions are created
        i = 0
        while (i < parameters["Tsamples"]):
            i +=1
            #create new solution
            new_solution = functions["changeSolution"]()(solution)
            #when already used in this temperature try to find new
            tries = 0
            while (new_solution in tabu_list) and (tries < parameters["Tsamples"]):
                new_solution = functions["changeSolution"]()(solution)
                tries += 1
            #append solution to tabu_list
            tabu_list.append(new_solution)
            #evaluate old and new solution
            old_cost = functions["costEvaluation"](solution)
            sum_old = old_cost[len(old_cost)-1]
            new_cost = functions["costEvaluation"](new_solution)
            sum_new = new_cost[len(new_cost)-1]
            #calculate difference
            d_cost = sum_new - sum_old
            #create random value
            randcrit = random.random()
            #if new solution better, then accept and save it
            if d_cost < 0:
                number_of_impr +=1
                solution = new_solution
                for cost in new_cost:
                    cost_data[new_cost.index(cost)].append(cost)
                cost_data[len(cost_data)-1].append(temp)
            #if new solution not better and not worse --> ignore it
            elif d_cost == 0:
                pass
            #if new solution worse, then
            else:
                #calculate accept probability
                exp = math.exp(-1*(d_cost*correct_factor/temp))
                #print exp
                #accept only with a probability
                if randcrit < exp and exp < 1.0:
                    number_of_acepted +=1
                    solution = new_solution
                    for cost in new_cost:
                        cost_data[new_cost.index(cost)].append(cost)
                    cost_data[len(cost_data)-1].append(temp)
        #end time estimation
        iend = time()
        #just for console check
        #print "Temp "+str(temp)
        #print "Improved "+str(number_of_impr)
        #print "Accepted "+str(number_of_acepted)
        #reduce temperature         
        temp *= alpha  
    #--------------------SAVING ROUTINE--------------------
    #stop psyco
    psyco.stop()
    #estimate needed time    
    iruntime = iend - istart
    #return time
    if time_check:
        parameters["Tsamples"] = tsamples
        parameters["Tsteps"] = tsteps
    if time_check:
        runtime =  tsteps*tsamples*iruntime/10
        return runtime
    #save files
    trial_name = file_info["trial_name"]
    file_name_list = file_info["file_name_list"]
    path = trial_name + '/Simulated Annealing/'
    if not os.path.exists(os.path.join(FOLDER_FOR_DATA, path)):
        os.makedirs(os.path.join(FOLDER_FOR_DATA, path))
    filename_list = []
    for list in cost_data:
        file_name_begin = file_name_list[cost_data.index(list)]
        i = 1
        filename = file_name_begin + ' %02i sta%.1f end%.1f ste%i sam%i.npy' % (i, parameters["Tstart"], parameters["Tend"], parameters["Tsteps"] ,parameters["Tsamples"])
        while os.path.exists(os.path.join(FOLDER_FOR_DATA, path, filename)):
            i += 1
            #print i
            filename = file_name_begin + ' %02i sta%.1f end%.1f ste%i sam%i.npy' % (i, parameters["Tstart"], parameters["Tend"], parameters["Tsteps"] ,parameters["Tsamples"])
        filename_list.append(os.path.join(path, filename))
        if not time_check:
            numpy.save(os.path.join(FOLDER_FOR_DATA, path, filename), numpy.float64(list))
    #return list of filenames    
    return filename_list