"""implements parser for benchmark files

Helper module implementing benchmark reading and creation capabilities

Right now, YAL and ACADEMIC benchmark support is implemented.

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>

@todo: When porting to python 3000 use abstract classes/methods
@todo: Decide if the entities of a benchmark are named blocks or modules
"""
from dsc_suite.tools.toolpaths import FOLDER_FOR_BENCHMARKS

##### BENCHMARK CLASS ##########################################################
class Benchmark(object):
    """Encapsulates basic benchmark operations.
    Use this class for benchmark format independent operations.
    Example: Use this class for random benchmark generation.

    This class currently handles only rectangular blocks (i.e.,
    storing only dimension informations of blocks). Benchmark
    status information is used to distinguish 2D/3D,
    rectangular/rectilinear, ...

    The most important variable is data, which stores the
    benchmark information. It's format is as following:
    data = {modulename : [moduletype, dimensions, connections]}
    where dimensions is a list and connections is a dictionary.
    
    @cvar MODULETYPE: Index of the module type
    @type MODULETYPE: int
    
    @cvar DIMENSIONS: Index of the dimensions
    @type DIMENSIONS: int
    
    @cvar CONNECTIONS: Index of the connections
    @type CONNECTIONS: int
    """
    # static members / constants
    MODULETYPE = 0
    DIMENSIONS = 1
    CONNECTIONS = 2
    # if something is changed here, also update the parsing methods
    #DATA = [MODULETYPE, DIMENSIONS, CONNECTIONS]
    
    def __init__(self, data, status):
        """Creates a new instance.
        Takes data and creates an own benchmark instance. data needs to
        be in the right format. No format testing is done!
        Example status = 
                          
        @param data: data from an external source
        data needs to be in right format:
            - {'source' : 'YAL', 'name' : filename, 'number of dimensions' : 2, 'shape of modules' : 'rectangluar'}
        @type data: dictionary
        
        @param status: benchmark status dictionary
        @type status: dictionary
        """
        self._status = status
        self.data = data
    
    @staticmethod
    def __abstract():
        """Static helper method to implement abstract methods"""
        import inspect
        caller = inspect.getouterframes(inspect.currentframe())[1][3]
        raise NotImplementedError(caller + ' must be implemented in subclass')
    
    @staticmethod
    def load_benchmark_file(filename):
        """Loads benchmark file and reads content into a single string for further processing.

        @param filename: filename of the benchmark file, path is constructed with default folder for benchmarks
        @type filename: string
        
        @return: Returns the content of benchmark file in a single string.
        @rtype: string
        """
        string = str()
        textfile = open(filename)
        for i in textfile:
            string += i
        textfile.close()
        return string

    def parse_benchmark_string(self, text):
        """Parses a benchmark string into the dictionary representation.
        Parses string representation of a SPECIFIC benchmark.
        data format: {modulename: [moduletype, dimensions, connections]}
        
        Abstract method!!! Needs to be implemented in subclass.

        @param text: string representation of benchmark file
        @type text: string
        
        @return: Returns the dictionary representation of a benchmark. 
        @rtype: dictionary
        """        
        self.__abstract()

    def load_benchmark(self, filename):
        """Loads a benchmark file. Combines load_benchmark_file and parse_benchmark_string.
        This method shall be executed in every __init__ of a file based
        benchmark (e.g., YAL). It loads the benchmark file and parses
        the content into the data variable, which then is accessible
        through several getter methods (e.g., get_blocks, get_dimensions)
        or by direct access (e.g., blocks, dimensions).

        @param filename: The name of the benchmark file, path is constructed with default folder for benchmarks.
        @type filename: string
        """
        text = self.load_benchmark_file(FOLDER_FOR_BENCHMARKS + '/' + filename)
        self.data = self.parse_benchmark_string(text)

    def get_status(self):
        """Reads the benchmark status dictionary.
        Getter method for the status dictionary.
        
        @return: Returns a dictionary containing status information regarding benchmark.
        keys:
            - source
            - name
            - number of dimensions
            - shapes of modules
        @rtype: dictionary
        """
        return self._status

    def get_blocks(self):
        """Reads the list of module names.
        Getter method for getting a list of module names.
        
        @return: Returns the list of module names.
        @rtype: list-of-strings
        """
        # Attention: returns a reference!
        return self.data.keys()

    def rename_blocks(self, block_list):
        """Renames the blocks in self.data without warnings.
        Renames blocks with the names in block_list.
        block_list format: [[a,b], [c,d], ...]
        Where a's new name would be b and c's new name d.
        If there was a b or before it will be overwritten
        without a warning.
        
        @param block_list: a list of block name tuples
        @type block_list: list-of-strings
        """
        # renaming blocks
        for old, new in block_list:
            if old in self.data:
                self.data[new] = self.data[old]
                del self.data[old]
        # renaming connections
        for block in self.data:
            for old, new in block_list:
                if old in self.data[block][self.CONNECTIONS]:
                    self.data[block][self.CONNECTIONS][new] = self.data[block][self.CONNECTIONS][old]
                    del self.data[block][self.CONNECTIONS][old]


    def remove_blocks(self, block_list):
        """Removes the blocks in self.data without warnings.
        Removes the blocks in block_list and their connections to other blocks.
        
        @param block_list: a list of block name tuples
        @type block_list: list-of-strings
        """
        for name in block_list:
            del self.data[name]
        for block in self.data:
            for name in block_list:
                if self.data[block][self.CONNECTIONS].has_key(name):
                    del self.data[block][self.CONNECTIONS][name]


    def get_dimensions(self):
        """Reads the module dimensions.
        Getter method to a dictionary of module dimensions.
        Dimensions are floats. Returns a dictionary with
        dimensions corresponding to len(data[X][DIMENSIONS])
        (i.e., 'number of dimensions').
        
        @return: Returns a dictionary of module dimensions.
        dimensions format: {'M1' : [X, Y, Z], ...}
        @rtype: dictionary
        """
        temp = dict()
        for i in self.data:
            # Attention: returns references!
            temp[i] = self.data[i][self.DIMENSIONS]
        return temp

    def get_total_area(self):
        assert(self._status['number of dimensions'] == 2)
        area = 0.0
        for i in self.data:
            # Attention: returns references!
            area += self.data[i][self.DIMENSIONS][0] * self.data[i][self.DIMENSIONS][1]
        return area

    
    def get_total_volume(self):
        assert(self._status['number of dimensions'] == 3)
        volume = 0.0
        for i in self.data:
            # Attention: returns references!
            volume += self.data[i][self.DIMENSIONS][0] * self.data[i][self.DIMENSIONS][1] * self.data[i][self.DIMENSIONS][2]
        return volume
        

    def get_connections(self):
        """Reads the module connections. 
        Getter method to a dictionary of module connections.
        Returns a dictionary with connections to other any other
        module. Maybe it could be optimized removing nonexisting
        (i.e., weight == 0) connections.
        @return: Returns a dictionary of the module connections.
        connections format: {'MN' : connection weight factor, ...}
        @rtype: dictionary
        """
        temp = dict()
        for i in self.data:
            # Attention: returns references!
            temp[i] = self.data[i][self.CONNECTIONS]
        return temp

    def join(self, benchmark): # vielleicht auch den + operator der Klasse ueberschreiben
        """Adds new benchmark data to the current benchmark.
        Add new benchmark data to the current benchmark. Identical module names
        are renamed. Only benchmarks with the same dimensionalty can be merged.
        The resulting benchmark don't have connections between the modules of
        the different benchmarks. So, typically further connections should be
        added (e.g., using the random connection generator of the RandomBench
        class).
        
        @param benchmark: a benchmark which is added to current data
        @type benchmark: Benchmark
        """
        from exceptions import ValueError
        if not (self.status['number of dimensions'] == benchmark.status['number of dimensions']):
            raise ValueError('joining benchmarks: number of dimensions does not match')

        self._status['source'] += '; ' + benchmark.status['source']
        self._status['name'] += '; ' + benchmark.status['name']

        if not (benchmark.status['shape of modules'] in self.status['shape of modules']):
            self._status['shape of modules'] += '; ' + benchmark.status['shape of modules']
        
        for module in benchmark.data:
            if module in self.blocks:
                i = 1
                while module+'-'+str(i) in self.blocks:
                    i += 1
                self.data[module+'-'+str(i)] = benchmark.data[module]
            else:
                self.data[module] = benchmark.data[module]                                                
    
    def set_connections(self, connections):
        """Resets connections in the data of the current benchmark.
                
        @param connections: a list which contains the information about the connections to be reseted
        @type connections: list
        """
        for block in self.data:
            if block in connections:
                self.data[block][self.CONNECTIONS] = connections[block]
                remove_list = []
                for new in self.data[block][self.CONNECTIONS]:
                    if not new in self.data:
                        remove_list.append(new)
                for wrong in remove_list:
                    del self.data[block][self.CONNECTIONS][wrong]
                        
                
    def add_connections(self, connections):
        """Adds new connections in the data of the current benchmark.
        
        @attention: Currently not implemented.
                
        @param connections: a list which contains the information about the connections to be added
        @type connections: list
        """
        # maybe using a static method from RandomBench, where a connection list
        # is derived from given block list
        pass

    def remove_connections(self, connections):
        """Removes connections in the data of the current benchmark.
        
        @attention: Currently not implemented.
                
        @param connections: a list which contains the information about the connections to be removed
        @type connections: list
        """
        pass
    
    def to_3d(self, z='half-minimum'):
        """Transforms a 2D benchmark into a 3D benchmark.
        
        @param z: If this parameter is set, it is the new z dimension of all modules.
        If this parameter is not set, the modules will be transformed with a the half of the minimum dimension of all modules.
        @type z: int
        
        @todo: More versatile methods needs to be implemented. (e.g., randomized z-dimension, folded 2d, ...)
        """
        assert z > 0
        #from exceptions import ValueError
        if not (self.status['number of dimensions'] == 2):
            raise ValueError('transformation into 3D only possible for benchmarks previously 2d')
        # different methods should be possible
        if (type(z) == int) or (type(z) == float):
            new_z = z
        elif z == 'half-minimum':
            new_z = min(min(self.dimensions.values(), key=lambda x: x[0])[0],
                        min(self.dimensions.values(), key=lambda x: x[1])[1])
            new_z /= float(2)
        else:
            raise ValueError('unconsidered parameter given')
        for module in self.data:
            self.data[module][self.DIMENSIONS].append(new_z)
        self._status['number of dimensions'] = 3
    
    # managed attributes (only works with python new style classes)
    # which means inheriting from object
    status = property(get_status)
    blocks = property(get_blocks)
    dimensions = property(get_dimensions)
    connections = property(get_connections, set_connections)
        
##### YAL BENCHMARK SUBCLASS ###################################################
class YAL(Benchmark):
    """Reads YAL benchmarks.
    Use this class to load a YAL file.
    This is a 2D benchmark format.
    Currently tested only for benchmarks from folders
    'Beasley and OKP' and 'MCNC'.
    """
    def __init__(self, filename):
        """Creates a new instance.
        Loads the YAL benchmark and creates the data set.

        @param filename: the filename of the benchmark file, path is constructed with default folder for benchmarks
        @type filename: string
        """
        self._status = {'source' : 'YAL',
                        'name' : filename,
                        'number of dimensions' : 2,
                        'shape of modules' : 'rectangluar'}
        self.load_benchmark(filename)
      
    def parse_benchmark_string(self, text):
        """Parses a benchmark string into the dictionary representation.
        Parses string representation of a SPECIFIC benchmark.
        data format: {modulename: [moduletype, dimensions, connections]}

        @param text: string representation of benchmark file
        @type text: string
        
        @return: Returns the dictionary representation of a benchmark. 
        @rtype: dictionary
        """         
        data = dict()
        searchpos = 0

        modpos = text.find('MODULE', searchpos)

        while modpos != -1:
            endmodpos = text.find('ENDMODULE', modpos)
            modulename = text[modpos+7:text.find(';', modpos)].strip()
            typepos = text.find('TYPE', modpos)
            moduletype = text[typepos+5:text.find(';', typepos)].strip()
            dimpos = text.find('DIMENSIONS', modpos)
            dims = text[dimpos+11:text.find(';', dimpos)].split()
            xdim = (max(float(dims[0]), float(dims[2]), float(dims[4]), float(dims[6]))
                    - min(float(dims[0]), float(dims[2]), float(dims[4]), float(dims[6])))
            assert xdim >= 0
            ydim = (max(float(dims[1]), float(dims[3]), float(dims[5]), float(dims[7]))
                    - min(float(dims[1]), float(dims[3]), float(dims[5]), float(dims[7])))
            assert ydim >= 0
            # gewonnene daten speichern
            data[modulename] = [moduletype, [xdim, ydim], []]
            # iolist erst mal ignorieren
            # spaeter notwendig, um herauszufinden, welche bloecke
            # aussen liegen sollten (die meisten netze zu den ausgangspins)
            netpos = text.find('NETWORK', modpos)
            endnetpos = text.find('ENDNETWORK', netpos)
            semicolon = text.find(';', netpos)
            nextsemicolon = text.find(';', semicolon+1)
            while nextsemicolon < endnetpos < endmodpos:
                network = text[semicolon+1:nextsemicolon].split()
                data[network[1]][self.CONNECTIONS] = network[2:]
                semicolon, nextsemicolon = nextsemicolon, text.find(';', nextsemicolon+1)

            # als letztes noch das naechste modul finden
            modpos = text.find('MODULE', endmodpos+10)

        # erst mal PARENT-Block ignorieren
        for i in data:
            if data[i][self.MODULETYPE] == 'PARENT':
                temp = i
        del(data[temp])

        # wenn VDD und GND raussollen, dann hier noch machen!
        # wenn netznamen wichtig sind, auskommentieren!
        self.transform(data)

        return data

    def transform(self, data):
        """Transforms connections from net based to a combination of priority plus partner information for each module.
        Transforms connections from net based to a combination of
        priority plus partner information for each module. (i.e.,
        ['net1', 'net2', ...] -> {block_x : 3, block_y : 4, ...})
        Not efficient, yet. Possible improvements: exploit symmetries,
        Modifies data dictionary given by reference!
        
        @param data: parsed benchmark data set
        @type data: dictionary.
        """
        temp = dict()
        for i in data:
            temp[i] = dict()
        for i in data:
            for j in data:
                # wie viele gemeinsame netze ?
                cn = len(set(data[i][self.CONNECTIONS]).intersection(set(data[j][self.CONNECTIONS])))
                temp[i][j] = cn
        for i in data:
            data[i][self.CONNECTIONS] = temp[i]


##### RANDOM BENCHMARK SUBCLASS ################################################
class RandomBench(Benchmark):

    """Generates random benchmark.
    Use this subclass to the basic Benchmark class to generate
    random benchmarks. Various approaches can be used (i.e.,
    different distributions for both, area and aspect ratio of
    generated modules). Generation of a 2D benchmark.
    """
    
    def __init__(self, n=10, method='uniform', aspect_ratio=(1,2), area=(10,100), ndigits=1):
        """Generates random benchmark with n modules.
        Generates random benchmark with n modules. For generation of random
        benchmarks a variety of approaches can be used (determined by given
        method). Area and aspect ratio of modules are limited by a tuple
        (e.g., (1.3, 2.5)).

        Following methods are possible:
            - 'uniform'     uniform distribution
            - 'gauss'       gauss distribution
            - 'half-gauss'  folded gauss distribution
            - 'exponential' exponetial distribution
        
        Combination of different methods possible to use a different
        distribution for aspect ratio and area, respectively.
        (e.g., 'uniform/gauss' -> uniform distribution for aspect ratio, gauss distribution for area)

        The meaning of the tuple (a, b) is as follows:
            - a: the lower/left limit
            - b: the upper/right limit

        @param n: number of generated modules
        @type n: int
        
        @param method: random generation method
        @type method: string
        
        @param aspect_ratio: tuple (a, b), parameterizing aspect ratio
        @type aspect_ratio: tuple
        
        @param area: tuple (a, b), parameterizing area
        @type area: tuple
        
        @param ndigits: number of digits used for rounding
        @type ndigits: int
        """
        status = {'source' : 'Random Benchmark',
                  'name' : "%s, %s, ar=%s, area=%s" % (str(n), method, str(aspect_ratio), str(area)),
                  'number of dimensions' : 2,
                  'shape of modules' : 'rectangluar'}
        dimensions = self.generate_dimensions(n, method, aspect_ratio, area, ndigits)
        #TODO connections = self.generate_connections(interconnect_method, interconnect_percentage ...
        # data format: {modulename: [moduletype, dimensions, connections]}
        data = dict()
        for module in dimensions:
            data[module] = ['GENERAL', dimensions[module], {}]
        Benchmark.__init__(self, data, status)
        

    @staticmethod
    def generate_dimensions(n, method, aspect_ratio, area, ndigits):
        """Generates random dimensions with n modules.
        Generates random dimensions with n modules. For generation of random
        benchmarks a variety of approaches can be used (determined by given
        method). Area and aspect ratio of modules are limited by a tuple
        (e.g., (1.3, 2.5)).

        Following methods are possible:
            - 'uniform'     uniform distribution
            - 'gauss'       gauss distribution
            - 'half-gauss'  folded gauss distribution
            - 'exponential' exponetial distribution
        
        Combination of different methods possible to use a different
        distribution for aspect ratio and area, respectively.
        (e.g., 'uniform/gauss' -> uniform distribution for aspect ratio, gauss distribution for area)
        The meaning of the tuple (a, b, c) is as follows:
            - a: the lower/left limit
            - b: the upper/right limit
            
        @todo: Depending on method, various parameters can be given for the
        distribution. Currently predetermined parameters are used.
            - 'uniform'          ... lower limit, upper limit
            - 'gauss/half-gauss' ... mu - mean value, sigma - standard deviation
            - 'exponential'      ... lambda

        FUTURE IDEA: Implementing a flag random, which determines, if the
        returned distributions are randomized or calculated (e.g., uniform
        can mean: uniform random distribution or real uniform distribution,
        such as range(1,10))

        2nd FUTURE IDEA: Generating dimensions by randomizing not aspect ratio
        and area, but width and height directly!!!
        
        
        
        @param n: number of generated modules
        @type n: int
        
        @param method: random generation method
        @type method: string
        
        @param aspect_ratio: tuple (a, b), parameterizing aspect ratio
        @type aspect_ratio: tuple
        
        @param area: tuple (a, b), parameterizing area
        @type area: tuple
        
        @param ndigits: number of digits used for rounding
        @type ndigits: int

        @return: Returns a ditionary with n random generated dimensions.
        @rtype: dictionary 
        """
        from exceptions import ValueError
        from dsc_suite.tools.combinatorics import EnumerativeCombinatorics
        from dsc_suite.tools.statistics import ProbabilityDistributions

        method_list = ProbabilityDistributions.IMPLEMENTED_DISTRIBUTIONS
        method_variations = EnumerativeCombinatorics.vary(method_list, 2)
        possible_methods = [m[0]+'/'+m[1] for m in method_variations] + method_list
        if not (method in possible_methods):
            raise ValueError('%s is not a valid distribution for RandomBench!' % method)
         
        if '/' in method:
            aspect_ratio_method, area_method = method.split('/')
        else:
            aspect_ratio_method = method
            area_method = method

        aspect_ratio_distribution = ProbabilityDistributions.get_distribution(n, aspect_ratio_method, *(aspect_ratio))
        area_distribution = ProbabilityDistributions.get_distribution(n, area_method, *(area))

        # maybe later given by parameter flag
        #aspect_ratio_distribution.sort()
        #area_distribution.sort()
        
        dimensions = dict()
        for i in range(n):
            width = (area_distribution[i]/aspect_ratio_distribution[i])**0.5
            height = area_distribution[i]/width 
            dimensions['M%02i' % i] = [round(width, ndigits), round(height, ndigits)]
        return dimensions
    
class ACADEMIC(Benchmark):
    """Reads the academic benchmarks.
    Use this class to load a ACADEMIC file.
    This is a 3D benchmark format.
    """
    def __init__(self, filename):
        """Creates a new instance.
        Loads the ACADEMIC benchmark file and creates the data set.
        @param filename: the filename of the benchmark file, path is constructed with default folder for benchmarks
        @type filename: string
        """
        self._status = {'name' : filename,
                        'number of dimensions' : 3,
                        'shape of modules' : 'rectangular',
                        'source' : 'BrainOfMine'}
        self.load_benchmark(filename)
            
    def parse_benchmark_string(self, text):
        """Parses a benchmark string into the dictionary representation.
        Parses string representation of a SPECIFIC benchmark.
        data format: {modulename: [moduletype, dimensions, connections]}

        @param text: string representation of benchmark file
        @type text: string
        
        @return: Returns the dictionary representation of a benchmark. 
        @rtype: dictionary
        """   
        data = eval(text)
        return data
    
class GSRC(Benchmark):
    """Reads GSRC benchmarks.
    Use this class to load GSRC files (.blocks, .nets).
    This is a 2D benchmark format.
    
    @attention: The seperation string used in the net list parsing could cause problems. Adjust if necessary.
    """
    def __init__(self, name):
        """Creates a new instance.
        Loads the YAL benchmark and creates the data set.

        @param name: the name of the benchmark, paths are constructed with default folder for benchmarks
        @type name: string
        """
        self._status = {'source' : 'GSRC',
                        'name' : name,
                        'number of dimensions' : 2,
                        'shape of modules' : 'rectangluar'}
        self.load_benchmark(name)
        
    def load_benchmark(self, name):
        """Loads the benchmark files. Combines load_benchmark_file and parse_benchmark_string.
        This method shall be executed in every __init__ of a file based
        benchmark (e.g. GSRC, YAL). It loads the benchmark file/s and parses
        the content into the data variable, which then is accessible
        through several getter methods (e.g., get_blocks, get_dimensions)
        or by direct access (e.g., blocks, dimensions).

        @param name: The name of the benchmark, paths are constructed with default folder for benchmarks.
        @type name: string
        """
        #getting information out of the .blocks and .nets files
        block_text = self.load_benchmark_file(FOLDER_FOR_BENCHMARKS + '/' + name + ".blocks")
        nets_text =  self.load_benchmark_file(FOLDER_FOR_BENCHMARKS + '/' + name + ".nets")
        self.data =  self.parse_benchmark_string(block_text, nets_text)
    
    def parse_benchmark_string(self, text_block, text_net):
        """Parses a benchmark string into the dictionary representation.
        Parses string representation of a SPECIFIC benchmark.
        data format: {modulename: [moduletype, dimensions, connections]}

        @param text_block: string representation of benchmark .blocks file
        @type text_block: string
        
        @param text_net: string representation of benchmark .nets file
        @type text_net: string
        
        @return: Returns the dictionary representation of a benchmark. 
        @rtype: dictionary
        """
        #save dict
        data = dict()
        
        #only hardrectilinear tested yet, no other files avaiable
        block_pos_h = text_block.find("hardrectilinear")
        block_pos_s = text_block.find("softrectangular")
        
        #getting index of block definitions
        if block_pos_h > 0 and block_pos_s > 0:
            if block_pos_h < block_pos_s:
                block_pos = block_pos_h
            else:
                block_pos = block_pos_s
        elif block_pos_h > 0 and block_pos_s < 0:
            block_pos = block_pos_h
        elif block_pos_h < 0 and block_pos_s > 0:
            block_pos = block_pos_s
        else:
            block_pos = -1
        #running through text    
        while block_pos != -1:
            #seperate line
            start = text_block.rfind("\n", 0, block_pos)
            end = text_block.find("\n", block_pos)
            line =  text_block[start+1:end]
            line = line.strip()
            #split line
            seperation = line.split(" ")
            #getting information from syntax
            block_name = seperation[0]
            block_type = seperation[1]
            block_length = 0.0
            block_width = 0.0
            #getting dimension information
            #only rectangle format supported here !
            open_bracket_pos = line.find("(")
            while open_bracket_pos != -1:
                close_bracket_pos = line.find(")",open_bracket_pos)
                pair = line[open_bracket_pos+1:close_bracket_pos]
                pair = pair.split(",")
                if float(pair[0]) > block_length:
                    block_length = float(pair[0])
                if float(pair[1]) > block_width:
                    block_width = float(pair[1])
                open_bracket_pos = line.find("(",close_bracket_pos)
            #writing into save dict with specific format
            data[block_name] = [block_type, [block_length, block_width], []]
            
            #getting index of next block definitions
            block_pos_h = text_block.find("hardrectilinear", end)
            block_pos_s = text_block.find("softrectangular", end)
            if block_pos_h > 0 and block_pos_s > 0:
                if block_pos_h < block_pos_s:
                    block_pos = block_pos_h
                else:
                    block_pos = block_pos_s
            elif block_pos_h > 0 and block_pos_s < 0:
                block_pos = block_pos_h
            elif block_pos_h < 0 and block_pos_s > 0:
                block_pos = block_pos_s
            else:
                block_pos = -1
                
        #getting terminal names 
        terminals = list()
        terminal_pos = text_block.find("terminal")
        while terminal_pos != -1:
            #seperate line
            start = text_block.rfind("\n", 0, terminal_pos)
            end = text_block.find("\n", terminal_pos)
            line =  text_block[start+1:end]
            line = line.strip()
            seperation = line.split(" ")
            #get terminal name from syntax
            terminals.append(seperation[0])
            terminal_pos = text_block.find("terminal", end)
        
        #reading net list
        net_pos = text_net.find("NetDegree")
        i = 0
        while net_pos != -1:
            #getting net list block
            start = text_net.find("\n", net_pos)
            net_pos = text_net.find("NetDegree", start)
            lines = text_net[start+1:net_pos]
            #seperate blocks from each other
            seperation = lines.split(" B")
            #naming net
            net_name = "net_%d" %i
            #search for terminal <--> net_name
            for k in range(0,len(seperation),1):
                seperation[k] = seperation[k].strip()
                if seperation[k] in terminals:
                    net_name = seperation[k]
            #update connection part in save dict for each block               
            for block in seperation:
                if net_name == block or block == "":
                    continue
                data[block][2].append(net_name)
            #avoid double net_names
            i +=1
        #transform connection style --> see class YAL
        self.transform(data)
        return data
    
    def transform(self, data):
        """Transforms connections from net based to a combination of priority plus partner information for each module.
        Transforms connections from net based to a combination of
        priority plus partner information for each module. (i.e.,
        ['net1', 'net2', ...] -> {block_x : 3, block_y : 4, ...})
        Not efficient, yet. Possible improvements: exploit symmetries,
        Modifies data dictionary given by reference!
        
        @param data: parsed benchmark data set
        @type data: dictionary.
        """
        temp = dict()
        for i in data:
            temp[i] = dict()
        for i in data:
            for j in data:
                # wie viele gemeinsame netze ?
                cn = len(set(data[i][self.CONNECTIONS]).intersection(set(data[j][self.CONNECTIONS])))
                temp[i][j] = cn
        for i in data:
            data[i][self.CONNECTIONS] = temp[i]
