"""implements several combinatoric basic methods

@author: Robert Fischbach
@author: Tobias Heimpold
@copyright: Copyright (C) 2011
@license: This program is free software; you can redistribute it and/or modify it 
          under the terms of the GNU General Public License as published by the 
          Free Software Foundation; either version 3 of the License, or (at your 
          option) any later version.
          This program is distributed in the hope that it will be useful, but 
          WITHOUT ANY WARRANTY; without even the implied warranty of 
          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
          Public License for more details.
          You should have received a copy of the GNU General Public License along 
          with this program; if not, see <http://www.gnu.org/licenses/>.
@contact: For questions and bugs send an email to <fischbach (at) ifte.de>
"""
from dsc_suite.tools.math import fak

##### ENUMERATIVE COMBINATORICS ################################################
class EnumerativeCombinatorics(object):
    """Implements basic EC functionality.
    This class implements basic enumerative combinatorics functionality, such
    as permutations, combinations and partitions, since they are necessary for
    my optimization investigations of data structures.

    Most methods are static (class) methods. Basically, this class is used to
    encapsulated conjoined functions.
    """

    #TODO: implement permutation method which allows to calculate
    #      the n-th permutation of an element list directly
    #      this would be a more memory saving solution for the data structures
    
    @staticmethod
    def permute_old(element_list):
        """Permutes all elements of the given list.
        Generates the PERMUTATION WITHOUT REPETITION of all elements given in
        the element_list, which are used once. The resulting permutation_list
        is a list including the n = len(element_list)! different sequences.
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @return: Returns all possible permutations of elements
        @rtype: list

        @attention: This permutation implementation delivers a different order
        than permute and get_permutation!
        """
        def p(permutation, elements, result):
            if len(elements) == 0:
                result.append(permutation)
            else:
                for e in elements:
                    temp = list(elements)
                    temp.remove(e)
                    if elements.index(e)%2 == 1:
                        temp.reverse()
                    temp2 = list(permutation)
                    temp2.append(e)
                    p(temp2, temp, result)
        
        permutation_list = []
        p([], list(element_list), permutation_list)
        return permutation_list

    @staticmethod
    def permute(element_list):
        """Permutes all elements of the given list.
        Generates the PERMUTATION WITHOUT REPETITION of all elements given in
        the element_list, which are used once. The resulting permutation_list
        is a list including the n = len(element_list)! different sequences.
        Obtains the same ordering of permutations as the get_permutation
        method.
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @return: Returns all possible permutations of elements
        @rtype: list
        """
        def p(processed, unprocessed):
            if len(unprocessed) == 1:
                return [processed + unprocessed]
            else:
                temp = []
                for i in range(len(unprocessed)):
                    temp.extend(p(processed + [unprocessed[i]], unprocessed[:i]+unprocessed[i+1:]))
                return temp
                
        permutation_list = p([], element_list)
        return permutation_list

    @staticmethod
    def get_permutation(element_list, number):
        """Permutes all elements of the given list.
        Generates the PERMUTATION WITHOUT REPETITION of the elements
        given in the element_list, which are used once. The number of
        all possible permutations is fak(len(element_list)), thus the
        range of number should be: 0 <= number < fak(len(element_list)).
        Validity of number is not tested. This should be done in the
        calling function (e.g., assert(number) < fak(len(element_list))),
        otherwise this test would be executed in every recursion step.
        
        Result equal to: permute(element_list)[number]
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @param number: number of permutation (implementation inherent order)
        @type number: int
        
        @return: Returns resulting permutation of elements.
        @rtype: list
        """
        if len(element_list) == 1:
            return element_list
        else:
            next_range = fak(len(element_list)-1)
            element_pos = number // next_range
            return [element_list[element_pos]] \
                + EnumerativeCombinatorics.get_permutation(element_list[:element_pos]+element_list[element_pos+1:],
                                                           number % next_range)
    
    @staticmethod
    def get_num_permutations(element_list):
        """Calculates number of possible permutations.
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @return: Returns the number of possible permutations.
        @rtype: int"""
        return fak(len(element_list)) # n!
    
    @staticmethod
    def get_num_variations(element_list, variation_length):
        """Calculates number of possible variations.
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @param variation_length: length of resulting variation
        @type variation_length: int
        
        @return: Returns the number of possible variations.
        @rtype: int"""
        return len(element_list)**variation_length # n^k
    
    @staticmethod
    def vary(element_list, variation_length):
        """Varies the elements in the given list.
        Generates the VARIATION WITH REPETITION of the given length. Therefore
        on every place every given element can be used. The order is important.
        Finaly, the variation_list contains len(element_list)**variation_length
        variations.
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @param variation_length: length of resulting variation
        @type variation_length: int
        
        @return: Returns the varied list.
        @rtype: list
        """
        if variation_length < 1:
            return []
        else:
            variation_list = [[x] for x in element_list]

        for i in range(variation_length-1):
            temp = []
            for element in element_list:
                for variation in variation_list:
                    temp.append([element] + variation)
            variation_list = temp
        
        return variation_list

    @staticmethod
    def get_variation(element_list, variation_length, number):
        """Generates variation and return the specified element.
        Generates one VARIATION WITH REPETITION of the given length and
        elements at position number.
        Result equal to: vary(element_list, variation_length)[number].
        
        @param element_list: list of elements which are used
        @type element_list: list
        
        @param variation_length: length of resulting variation
        @type variation_length: int
        
        @param number: number of wanted variation
        @type number: int 
        
        @return: Returns the wanted variation.
        @rtype: list
        """
        base = len(element_list)
        variation = []
        for i in range(variation_length-1, -1, -1):
            variation.append(element_list[number//base**i])
            number = number % base**i
        return variation
    
    #NEXT STEP: combination, variation, permutation with/without repetition
